# coding:utf-8
'''
Test Module's `.todevice` interface
According to [Issue](https://github.com/Theano/Theano/issues/6655), this feature of Theano is never finished.
Created   :  11,  2, 2018
Revised   :  11,  2, 2018
All rights reserved
'''
# ------------------------------------------------------------------------------------------------
__author__ = 'dawei.leng'
import os
os.environ['THEANO_FLAGS'] = "floatX=float32,mode=FAST_RUN, warn_float64='raise',contexts=dev0->cuda3;dev1->cuda6"
import theano
theano_path = os.path.split(theano.__file__)[0]
print('theano path = %s\n' % theano_path)
import theano.tensor as tensor
import dandelion
dandelion_path = os.path.split(dandelion.__file__)[0]
print('dandelion path = %s\n' % dandelion_path)
from dandelion.module import *


class build_model_on_single_device(Module):
    def __init__(self, in_dim=1024, out_dim=512, device_context=None):
        super().__init__()
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.dense1 = Dense(input_dims=self.in_dim, output_dim=self.out_dim)
        self.dense2 = Dense(input_dims=self.out_dim, output_dim=self.out_dim)
        if device_context is not None:
            self.dense1.todevice(device_context)
            self.dense2.todevice(device_context)

    def forward(self, x):
        x = self.dense1.forward(x)
        x = relu(x)
        x = self.dense2.forward(x)
        x = relu(x)
        return x

    def predict(self, x):
        x = self.dense1.predict(x)
        x = relu(x)
        x = self.dense2.predict(x)
        x = relu(x)
        return x

class build_model_on_multiple_device(Module):
    def __init__(self, in_dim=1024, out_dim=512, device_context=None):
        super().__init__()
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.dense1 = Dense(input_dims=self.in_dim, output_dim=self.out_dim)
        self.dense2 = Dense(input_dims=self.out_dim, output_dim=self.out_dim)
        if device_context is not None:
            self.dense1.todevice(device_context[0])
            self.dense2.todevice(device_context[1])
        self.device_context = device_context

    def forward(self, x):
        x = self.dense1.forward(x)
        x = relu(x)
        if self.device_context is not None:
            x = x.transfer(self.device_context[1])
        x = self.dense2.forward(x)
        x = relu(x)
        return x

    def predict(self, x):
        x = self.dense1.predict(x)
        x = relu(x)
        if self.device_context is not None:
            x = x.transfer(self.device_context[1])
        x = self.dense2.predict(x)
        x = relu(x)
        return x

def test_case_0(batch=256, in_dim=1024, out_dim=512):
    import numpy as np
    import time
    try:
        model_on_single_device   = build_model_on_single_device(in_dim=in_dim, out_dim=out_dim, device_context='dev0')
        model_on_multiple_device = build_model_on_multiple_device(in_dim=in_dim, out_dim=out_dim, device_context=['dev0', 'dev1'])
    except ValueError as e:
        if str(e).startswith("Can't transfer to target"):
            print('GPU not present, test skipped')
            return

    model_on_multiple_device.dense1.W.set_value(model_on_single_device.dense1.W.get_value())
    model_on_multiple_device.dense1.b.set_value(model_on_single_device.dense1.b.get_value())
    model_on_multiple_device.dense2.W.set_value(model_on_single_device.dense2.W.get_value())
    model_on_multiple_device.dense2.b.set_value(model_on_single_device.dense2.b.get_value())

    x = tensor.fmatrix()
    x1 = x.transfer('dev0')
    y0 = model_on_single_device.forward(x1)
    y1 = model_on_multiple_device.forward(x1)

    f0 = theano.function([x], y0, no_default_updates=True)
    f1 = theano.function([x], y1, no_default_updates=True)

    for i in range(20):
        x = np.random.rand(batch, in_dim).astype(np.float32)
        time00 = time.time()
        y0 = f0(x)
        time01 = time.time()
        y1 = f1(x)
        time02 = time.time()
        time0 = time01 - time00
        time1 = time02 - time01
        diff = np.sum(np.abs(y0 - y1))
        print('i=%d, diff=%0.6f, time0=%0.6fs, time1=%0.6fs, time0/time1=%0.4f' % (i, diff, time0, time1, time0/time1))
        if diff>1e-4:
            raise ValueError('diff is too big')

if __name__ == '__main__':
    try:
        test_case_0(batch=512, in_dim=1024, out_dim=512)
    except ValueError as e:
        if str(e).startswith("Can't transfer to target"):
            print('GPU not present, test skipped')
    print('Test passed')


# coding:utf-8
'''
  Function set for image processing and computer vision
  Created   :  10, 19, 2015
  Revised   :   4,  4, 2018  total rewrite of `imread()` to add support of EXIF rotation handling
                8, 16, 2018  add `border_mode` arg to `imrotate()`, the `interpolation` arg type is changed to string.
  All rights reserved
'''
__author__ = 'dawei.leng'

import numpy as np
from PIL import Image, ImageFile, ExifTags
ImageFile.LOAD_TRUNCATED_IMAGES = True

def imread(f, flatten=False, dtype='float32'):
    """
    Read image from file, backend=scipy
    f       : str or file object -> The file name or file object to be read.
    flatten : bool, optional  -> If True, flattens the color channels into a single gray-scale channel.
    Returns ndarray
    """
    # return sp.misc.imread(f, flatten)

    image=Image.open(f)
    
    try:
        # --- handle EXIF rotation ---#
        for orientation in ExifTags.TAGS.keys() :
            if ExifTags.TAGS[orientation]=='Orientation' : break
        exif=dict(image._getexif().items())
        if   exif[orientation] == 3 :
            image=image.rotate(180, expand=True)
        elif exif[orientation] == 6 :
            image=image.rotate(270, expand=True)
        elif exif[orientation] == 8 :
            image=image.rotate(90, expand=True)
    finally:
        #--- return proper numpy array ---#
        if dtype == 'uint8':
            if flatten:
                return np.array(image.convert('L'))
            else:
                return np.array(image.convert('RGB'))
        else: # dtype == 'float32'
            if flatten:
                return np.array(image.convert('F'))
            else:
                return np.array(image.convert('RGB'), dtype='float32')

def imsave(f, I, **params):
    """
    Save image into file, backend=pillow
    For jpeg, image should be in uint8 type
    :param f: str or ﬁle object
    :param I:
    :return:
    """
    image = Image.fromarray(I)
    return image.save(fp=f, format=None, **params)

def imresize(I, size, interp='bilinear', mode=None):
    import scipy.misc
    return scipy.misc.imresize(I, size, interp, mode)

def imrotate(I, angle, padvalue=0.0, interpolation='linear', target_size=None, border_mode='reflect_101'):
    """
    Return a rotated image, backend=opencv
    :param I:  N-D np array, dtype not limited
    :param angle:   in degree, positive for counter-clockwise
    :param border_mode: image boundary handling method, {'reflect_101'|'reflect'|'wrap'|'constant'|'replicate'}, refer to opencv:BORDER_* constants for details
    :param padvalue: used when `border_mode` = 'constant'
    :param interpolation: image interpolation method, {'linear'|'nearest'|'cubic'|'LANCZOS4'|'area'}, refer to opencv:INTER_* constants for details
    :return: rotated image, dtype same with `I`
    """
    import cv2
    cv2.INTER_LINEAR
    assert border_mode.lower() in {'reflect_101', 'reflect', 'wrap', 'constant', 'replicate'}
    assert interpolation.lower() in {'linear', 'nearest', 'cubic', 'lanczos4', 'area'}
    if abs(angle) < 0.01:
        return I
    rows, cols = I.shape[:2]
    rmatrix = cv2.getRotationMatrix2D((cols/2, rows/2), angle, 1.0)                       # 2 * 3
    vertex_matrix = np.array([[0, 0, 1], [cols, 0, 1], [0, rows, 1], [cols, rows, 1]]).T  # 3 * 4
    vertex_matrix_target = np.dot(rmatrix, vertex_matrix)                                 # 2 * 4
    rowmin, colmin = vertex_matrix_target.min(axis=1)
    rowmax, colmax = vertex_matrix_target.max(axis=1)
    newshape = (rowmax-rowmin, colmax-colmin)
    rmatrix[0, 2] += newshape[0]/2 - cols/2
    rmatrix[1, 2] += newshape[1]/2 - rows/2
    if target_size is not None:
        newshape = target_size
    else:
        newshape = (np.round(newshape[0]).astype('int'), np.round(newshape[1]).astype('int'))
    I2 = cv2.warpAffine(I, rmatrix, newshape,
                        flags=getattr(cv2, 'INTER_%s' % interpolation.upper()),
                        borderMode=getattr(cv2, 'BORDER_%s' % border_mode.upper()),
                        borderValue=padvalue)
    return I2


if __name__ == '__main__':
    from matplotlib import pyplot as plt
    file = r"C:\Users\dawei\Work\Data\Robust_Reading\ICDAR2013_robust_reading_challenge2\trainset\images\107.jpg"
    I = imread(file, flatten=False)
    I2 = imrotate(I, 5, border_mode='reflect_101', interpolation='linear')
    plt.imshow(I2.astype('uint8'), 'gray')
    plt.show()



# coding:utf-8
# Test for Sequential container
# Created   :   7, 12, 2018
# Revised   :   7, 12, 2018
# All rights reserved
#------------------------------------------------------------------------------------------------
__author__ = 'dawei.leng'
import os
os.environ['THEANO_FLAGS'] = "floatX=float32, mode=FAST_RUN, warn_float64='raise'"

import theano
from theano import tensor
from dandelion.module import *
from dandelion.activation import *

import dandelion
dandelion_path = os.path.split(dandelion.__file__)[0]
print('dandelion path = %s\n' % dandelion_path)

def test_case_0():
    conv1 = Conv2D(in_channels=3, out_channels=3, stride=(2, 2))
    bn1   = BatchNorm(input_shape=(None, 3, None, None))
    conv2 = Conv2D(in_channels=3, out_channels=5)
    conv3 = Conv2D(in_channels=5, out_channels=8)
    model = Sequential([conv1, bn1, conv2, conv3], activation=relu, name='seq')
    model_weights = model.get_weights()
    for value, w_name in model_weights:
        print('name = %s, shape=' % w_name, value.shape)

    x = tensor.ftensor4('x')
    y = model.forward(x)
    print('compiling fn...')
    fn = theano.function([x], y, no_default_updates=False)
    print('run fn...')
    input = np.random.rand(4, 3, 32, 33).astype(np.float32)
    output = fn(input)
    print(output)
    print(output.shape)

if __name__ == '__main__':

    test_case_0()

    print('Test passed')






# coding:utf-8
"""
LSTM2D implementation by alternating LSTM along different dimensions
Created   :  11,  9, 2018
Revised   :  11,  9, 2018
All rights reserved
"""
# ------------------------------------------------------------------------------------------------
__author__ = 'dawei.leng'

import theano.tensor as tensor
from ..module import *
from ..functional import *
from ..activation import *

class Alternate_2D_LSTM(Module):
    """
    2D LSTM implementaton by alternating 1D LSTM along different input dimensions
    Input shape = (H, W, B, C)

    """
    def __init__(self, input_dims, hidden_dim, peephole=True, initializer=init.Normal(0.1), grad_clipping=0, hidden_activation=tanh,
                 learn_ini=False, truncate_gradient=-1, mode=2):
        """
        :param input_dims: integer or list of integers, dimension of the input for different part, allows to have different
                            initializations for different parts of the input.
        :param hidden_dim:
        :param initializer:
        :param peephole: whether add peephole connection
        :param grad_clipping: float. Hard clip the gradients at each time step. Only the gradient values
                              above this threshold are clipped to the threshold. This is done during backprop.
        :param hidden_activation: nonlinearity applied to hidden variable, i.e., h = out_gate * hidden_activation(cell). It's recommended to use `tanh` as default.
        :param learn_ini: If True, initial hidden values will be learned.
        :param truncate_gradient: if not -1, BPTT will be used, gradient back-propagation will be performed at most `truncate_gradient` steps
        :param mode: {0|1|2}. 0: 1D LSTM results from horizontal and vertical dimensions are concatenated along the `C` dimension; 1: horizontal and vertical dimensions are processed
               sequentially, i.e., the final result = horizontal_LSTM(vertical_LSTM(input)); 2: mixed mode: final result = horizontal_LSTM(concat(input, vertical_LSTM(input)))

        """
        super().__init__()
        self.hidden_dim = hidden_dim
        self.mode = mode
        if not isinstance(input_dims, (tuple, list)):
            input_dims = [input_dims]
        self.lstm_h = LSTM(input_dims=input_dims, hidden_dim=hidden_dim, peephole=peephole, initializer=initializer, grad_clipping=grad_clipping, hidden_activation=hidden_activation,
                           learn_ini=learn_ini, truncate_gradient=truncate_gradient)
        if self.mode == 0:
            lstm_w_inputdim = input_dims
        elif self.mode == 1:
            lstm_w_inputdim = hidden_dim
        elif self.mode == 2:
            lstm_w_inputdim = input_dims + [hidden_dim]
        else:
            raise ValueError('Invalid mode input: should be among {0, 1, 2}')
        self.lstm_w = LSTM(input_dims=lstm_w_inputdim, hidden_dim=hidden_dim, peephole=peephole, initializer=initializer, grad_clipping=grad_clipping, hidden_activation=hidden_activation,
                           learn_ini=learn_ini, truncate_gradient=truncate_gradient)
        self.predict = self.forward

    def forward(self, seq_input, h_ini=(None, None), c_ini=(None, None), seq_mask=None, backward=(False, False), return_final_state=False):
        """
        :param seq_input:    (H, W, B, input_dim)
        :param h_ini:        tuple of matrix (B, hidden_dim) or None, if None, then learned self.h_ini will be used
        :param c_ini:        tuple of matrix (B, hidden_dim) or None, if None, then learned self.c_ini will be used
        :param seq_mask:     (H, W, B)
        :param backward:     tuple of False/True
        :param only_return_final:  If True, only return the final sequential output
        :param return_final_state: If True, the final state of `hidden` and `cell` will be returned, both (B, hidden_dim)
        :return: (H, W, B, hidden_dim) if mode= 1 or 2, (H, W, B, 2*hidden_dim) if mode = 0
        """
        H, W, B, C = seq_input.shape
        h_ini, c_ini, backward = as_tuple(h_ini, 2), as_tuple(c_ini, 2), as_tuple(backward, 2)
        x = tensor.reshape(seq_input, (H, W*B, C))
        output_h = self.lstm_h.forward(x, h_ini=h_ini[0], c_ini=c_ini[0], seq_mask=seq_mask, backward=backward[0], return_final_state=return_final_state)

        if self.mode == 0:
            x = seq_input.dimshuffle(1, 0, 2, 3)  # (W, H, B, C)
            x = tensor.reshape(x, (W, H*B, C))
            output_w = self.lstm_w.forward(x, h_ini=h_ini[1], c_ini=c_ini[1], seq_mask=seq_mask, backward=backward[1], return_final_state=return_final_state)
            if return_final_state:
                x_h, x_w = output_h[0], output_w[0]
            else:
                x_h, x_w = output_h, output_w
            x_h, x_w = tensor.reshape(x_h, (H, W, B, -1)), tensor.reshape(x_w, (W, H, B, -1))
            x_w = x_w.dimshuffle(1, 0, 2, 3)
            x = tensor.concatenate([x_h, x_w], axis=3)
            if return_final_state:
                return x, output_h[1:], output_w[1:]
            else:
                return x

        elif self.mode == 1:
            if return_final_state:
                x = output_h[0]  # (H, W*B, hidden_dim)
            else:
                x = output_h
            x = tensor.reshape(x, (H, W, B, self.hidden_dim))
            x = x.dimshuffle(1, 0, 2, 3)
            x = tensor.reshape(x, (W, H*B, self.hidden_dim))
            output_w = self.lstm_w.forward(x, h_ini=h_ini[1], c_ini=c_ini[1], seq_mask=seq_mask, backward=backward[1], return_final_state=return_final_state)
            if return_final_state:
                x = tensor.reshape(output_w[0], (W, H, B, -1))
                x = x.dimshuffle(1, 0, 2, 3)  # (H, W, B, hidden_dim)
                return x, output_h[1:], output_w[1:]
            else:
                x = tensor.reshape(output_w, (W, H, B, -1))
                x = x.dimshuffle(1, 0, 2, 3)  # (H, W, B, hidden_dim)
                return x

        else:
            if return_final_state:
                x = output_h[0]  # (H, W*B, hidden_dim)
            else:
                x = output_h
            x = tensor.reshape(x, (H, W, B, self.hidden_dim))  # (H, W, B, hidden_dim)
            x = tensor.concatenate([seq_input, x], axis=3)
            x = x.dimshuffle(1, 0, 2, 3)
            x = tensor.reshape(x, (W, H*B, -1))
            output_w = self.lstm_w.forward(x, h_ini=h_ini[1], c_ini=c_ini[1], seq_mask=seq_mask, backward=backward[1], return_final_state=return_final_state)
            if return_final_state:
                x = tensor.reshape(output_w[0], (W, H, B, -1))
                x = x.dimshuffle(1, 0, 2, 3)  # (H, W, B, hidden_dim)
                return x, output_h[1:], output_w[1:]
            else:
                x = tensor.reshape(output_w, (W, H, B, -1))
                x = x.dimshuffle(1, 0, 2, 3)  # (H, W, B, hidden_dim)
                return x





"""
Functions to create initializers for parameter variables.
"""

import numpy as np

from .util import floatX

class Initializer(object):
    """Base class for parameter tensor initializers.

    The :class:`Initializer` class represents a weight initializer used
    to initialize weight parameters in a neural network layer. It should be
    subclassed when implementing new types of weight initializers.

    """
    def __call__(self, shape):
        """
        Makes :class:`Initializer` instances callable like a function, invoking
        their :meth:`sample()` method.
        """
        return self.sample(shape)

    def sample(self, shape):
        """
        Sample should return a theano.tensor of size shape and data type
        theano.config.floatX.

        Parameters
        -----------
        shape : tuple or int
            Integer or tuple specifying the size of the returned
            matrix.
        returns : theano.tensor
            Matrix of size shape and dtype theano.config.floatX.
        """
        raise NotImplementedError()


class Normal(Initializer):
    """Sample initial weights from the Gaussian distribution.

    Initial weight parameters are sampled from N(mean, std).

    Parameters
    ----------
    std : float
        Std of initial parameters.
    mean : float
        Mean of initial parameters.
    """
    def __init__(self, std=0.01, mean=0.0):
        self.std = std
        self.mean = mean

    def sample(self, shape):
        return floatX(np.random.normal(self.mean, self.std, size=shape))


class Uniform(Initializer):
    """Sample initial weights from the uniform distribution.

    Parameters are sampled from U(a, b).

    Parameters
    ----------
    range : float or tuple
        When std is None then range determines a, b. If range is a float the
        weights are sampled from U(-range, range). If range is a tuple the
        weights are sampled from U(range[0], range[1]).
    std : float or None
        If std is a float then the weights are sampled from
        U(mean - np.sqrt(3) * std, mean + np.sqrt(3) * std).
    mean : float
        see std for description.
    """
    def __init__(self, range=0.01, std=None, mean=0.0):
        if std is not None:
            a = mean - np.sqrt(3) * std
            b = mean + np.sqrt(3) * std
        else:
            try:
                a, b = range  # range is a tuple
            except TypeError:
                a, b = -range, range  # range is a number

        self.range = (a, b)

    def sample(self, shape):
        return floatX(np.random.uniform(
            low=self.range[0], high=self.range[1], size=shape))


class Glorot(Initializer):
    """Glorot weight initialization.

    This is also known as Xavier initialization [1]_.

    Parameters
    ----------
    initializer : lasagne.init.Initializer
        Initializer used to sample the weights, must accept `std` in its
        constructor to sample from a distribution with a given standard
        deviation.
    gain : float or 'relu'
        Scaling factor for the weights. Set this to ``1.0`` for linear and
        sigmoid units, to 'relu' or ``sqrt(2)`` for rectified linear units, and
        to ``sqrt(2/(1+alpha**2))`` for leaky rectified linear units with
        leakiness ``alpha``. Other transfer functions may need different
        factors.
    c01b : bool
        For a :class:`lasagne.layers.cuda_convnet.Conv2DCCLayer` constructed
        with ``dimshuffle=False``, `c01b` must be set to ``True`` to compute
        the correct fan-in and fan-out.

    References
    ----------
    .. [1] Xavier Glorot and Yoshua Bengio (2010):
           Understanding the difficulty of training deep feedforward neural
           networks. International conference on artificial intelligence and
           statistics.

    Notes
    -----
    For a :class:`DenseLayer <lasagne.layers.DenseLayer>`, if ``gain='relu'``
    and ``initializer=Uniform``, the weights are initialized as

    .. math::
       a &= \\sqrt{\\frac{12}{fan_{in}+fan_{out}}}\\\\
       W &\sim U[-a, a]

    If ``gain=1`` and ``initializer=Normal``, the weights are initialized as

    .. math::
       \\sigma &= \\sqrt{\\frac{2}{fan_{in}+fan_{out}}}\\\\
       W &\sim N(0, \\sigma)

    See Also
    --------
    GlorotNormal  : Shortcut with Gaussian initializer.
    GlorotUniform : Shortcut with uniform initializer.
    """
    def __init__(self, initializer, gain=1.0, c01b=False):
        if gain == 'relu':
            gain = np.sqrt(2)

        self.initializer = initializer
        self.gain = gain
        self.c01b = c01b

    def sample(self, shape):
        if self.c01b:
            if len(shape) != 4:
                raise RuntimeError(
                    "If c01b is True, only shapes of length 4 are accepted")

            n1, n2 = shape[0], shape[3]
            receptive_field_size = shape[1] * shape[2]
        else:
            if len(shape) < 2:
                raise RuntimeError(
                    "This initializer only works with shapes of length >= 2")

            n1, n2 = shape[:2]
            receptive_field_size = np.prod(shape[2:])

        std = self.gain * np.sqrt(2.0 / ((n1 + n2) * receptive_field_size))
        return self.initializer(std=std).sample(shape)


class GlorotNormal(Glorot):
    """Glorot with weights sampled from the Normal distribution.

    See :class:`Glorot` for a description of the parameters.
    """
    def __init__(self, gain=1.0, c01b=False):
        super(GlorotNormal, self).__init__(Normal, gain, c01b)


class GlorotUniform(Glorot):
    """Glorot with weights sampled from the Uniform distribution.

    See :class:`Glorot` for a description of the parameters.
    """
    def __init__(self, gain=1.0, c01b=False):
        super(GlorotUniform, self).__init__(Uniform, gain, c01b)


class He(Initializer):
    """He weight initialization.

    Weights are initialized with a standard deviation of
    :math:`\\sigma = gain \\sqrt{\\frac{1}{fan_{in}}}` [1]_.

    Parameters
    ----------
    initializer : lasagne.init.Initializer
        Initializer used to sample the weights, must accept `std` in its
        constructor to sample from a distribution with a given standard
        deviation.
    gain : float or 'relu'
        Scaling factor for the weights. Set this to ``1.0`` for linear and
        sigmoid units, to 'relu' or ``sqrt(2)`` for rectified linear units, and
        to ``sqrt(2/(1+alpha**2))`` for leaky rectified linear units with
        leakiness ``alpha``. Other transfer functions may need different
        factors.
    c01b : bool
        For a :class:`lasagne.layers.cuda_convnet.Conv2DCCLayer` constructed
        with ``dimshuffle=False``, `c01b` must be set to ``True`` to compute
        the correct fan-in and fan-out.

    References
    ----------
    .. [1] Kaiming He et al. (2015):
           Delving deep into rectifiers: Surpassing human-level performance on
           imagenet classification. arXiv preprint arXiv:1502.01852.

    See Also
    ----------
    HeNormal  : Shortcut with Gaussian initializer.
    HeUniform : Shortcut with uniform initializer.
    """
    def __init__(self, initializer, gain=1.0, c01b=False):
        if gain == 'relu':
            gain = np.sqrt(2)

        self.initializer = initializer
        self.gain = gain
        self.c01b = c01b

    def sample(self, shape):
        if self.c01b:
            if len(shape) != 4:
                raise RuntimeError(
                    "If c01b is True, only shapes of length 4 are accepted")

            fan_in = np.prod(shape[:3])
        else:
            if len(shape) == 2:
                fan_in = shape[0]
            elif len(shape) > 2:
                fan_in = np.prod(shape[1:])
            else:
                raise RuntimeError(
                    "This initializer only works with shapes of length >= 2")

        std = self.gain * np.sqrt(1.0 / fan_in)
        return self.initializer(std=std).sample(shape)


class HeNormal(He):
    """He initializer with weights sampled from the Normal distribution.

    See :class:`He` for a description of the parameters.
    """
    def __init__(self, gain=1.0, c01b=False):
        super(HeNormal, self).__init__(Normal, gain, c01b)


class HeUniform(He):
    """He initializer with weights sampled from the Uniform distribution.

    See :class:`He` for a description of the parameters.
    """
    def __init__(self, gain=1.0, c01b=False):
        super(HeUniform, self).__init__(Uniform, gain, c01b)


class Constant(Initializer):
    """Initialize weights with constant value.

    Parameters
    ----------
     val : float
        Constant value for weights.
    """
    def __init__(self, val=0.0):
        self.val = val

    def sample(self, shape):
        return floatX(np.ones(shape) * self.val)


class Sparse(Initializer):
    """Initialize weights as sparse matrix.

    Parameters
    ----------
    sparsity : float
        Exact fraction of non-zero values per column. Larger values give less
        sparsity.
    std : float
        Non-zero weights are sampled from N(0, std).
    """
    def __init__(self, sparsity=0.1, std=0.01):
        self.sparsity = sparsity
        self.std = std

    def sample(self, shape):
        if len(shape) != 2:
            raise RuntimeError(
                "sparse initializer only works with shapes of length 2")

        w = floatX(np.zeros(shape))
        n_inputs, n_outputs = shape
        size = int(self.sparsity * n_inputs)  # fraction of number of inputs

        for k in range(n_outputs):
            indices = np.arange(n_inputs)
            np.random.shuffle(indices)
            indices = indices[:size]
            values = floatX(np.random.normal(0.0, self.std, size=size))
            w[indices, k] = values

        return w


class Orthogonal(Initializer):
    """Intialize weights as Orthogonal matrix.

    Orthogonal matrix initialization [1]_. For n-dimensional shapes where
    n > 2, the n-1 trailing axes are flattened. For convolutional layers, this
    corresponds to the fan-in, so this makes the initialization usable for
    both dense and convolutional layers.

    Parameters
    ----------
    gain : float or 'relu'
        Scaling factor for the weights. Set this to ``1.0`` for linear and
        sigmoid units, to 'relu' or ``sqrt(2)`` for rectified linear units, and
        to ``sqrt(2/(1+alpha**2))`` for leaky rectified linear units with
        leakiness ``alpha``. Other transfer functions may need different
        factors.

    References
    ----------
    .. [1] Saxe, Andrew M., James L. McClelland, and Surya Ganguli.
           "Exact solutions to the nonlinear dynamics of learning in deep
           linear neural networks." arXiv preprint arXiv:1312.6120 (2013).
    """
    def __init__(self, gain=1.0):
        if gain == 'relu':
            gain = np.sqrt(2)

        self.gain = gain

    def sample(self, shape):
        if len(shape) < 2:
            raise RuntimeError("Only shapes of length 2 or more are "
                               "supported.")

        flat_shape = (shape[0], np.prod(shape[1:]))
        a = np.random.normal(0.0, 1.0, flat_shape)
        u, _, v = np.linalg.svd(a, full_matrices=False)
        # pick the one with the correct shape
        q = u if u.shape == flat_shape else v
        q = q.reshape(shape)
        return floatX(self.gain * q)



# coding:utf-8
# Test for categorical_crossentropy_log
# Created   :   7,  3, 2018
# Revised   :   7,  3, 2018
# All rights reserved
#------------------------------------------------------------------------------------------------
__author__ = 'dawei.leng'
import os, numpy as np
os.environ['THEANO_FLAGS'] = "floatX=float32, mode=FAST_RUN, warn_float64='raise'"

import theano
from theano import tensor
from dandelion.module import *
from dandelion.activation import softmax, log_softmax
from dandelion.objective import categorical_crossentropy, categorical_crossentropy_log
from dandelion.util import one_hot

import dandelion
dandelion_path = os.path.split(dandelion.__file__)[0]
print('dandelion path = %s\n' % dandelion_path)

def test_case_0():
    x = tensor.fmatrix()  # (B, N)
    # y = tensor.ivector()  # (B,)
    y = tensor.fmatrix()  # (B, N)

    x1 = softmax(x)  # (B, N)
    r1 = categorical_crossentropy(x1, y, eps=0.0)

    x2 = log_softmax(x)
    r2 = categorical_crossentropy_log(x2, y)

    f1 = theano.function([x, y], r1)
    f2 = theano.function([x, y], r2)

    for i in range(100):
        B, N = np.random.randint(low=1, high=32), np.random.randint(low=2, high=100)
        X = np.random.rand(B, N)
        Y = np.random.randint(low=0, high=N, size=(B,))
        Y = np.eye(N)[Y]

        X = X.astype('float32')
        Y = Y.astype('float32')

        r1 = f1(X, Y)
        r2 = f2(X, Y)
        dif = np.sum(np.abs(r1 - r2))
        if dif > 1e-7:
            print(r1)
            print(r2)
            raise ValueError('r1 != r2')

def test_case_1():
    N = np.random.randint(low=2, high=100)
    x = tensor.fmatrix()  # (B, N)
    y = tensor.ivector()  # (B,)

    x1 = softmax(x)  # (B, N)
    r1 = categorical_crossentropy(x1, y, eps=0.0)

    x2 = log_softmax(x)
    r2 = categorical_crossentropy_log(x2, y, m=N)

    f1 = theano.function([x, y], r1)
    f2 = theano.function([x, y], r2)

    for i in range(100):
        # B, N = np.random.randint(low=1, high=32), np.random.randint(low=2, high=100)
        B = np.random.randint(low=1, high=32)
        X = np.random.rand(B, N)
        Y = np.random.randint(low=0, high=N, size=(B,))

        X = X.astype('float32')
        Y = Y.astype('int32')

        r1 = f1(X, Y)
        # print('r1=', r1)
        r2 = f2(X, Y)
        # print('r2=', r2)
        dif = np.max(np.abs(r1 - r2))
        if dif > 1e-6:
            print('r1=', r1)
            print('r2=', r2)
            raise ValueError('r1 != r2')

if __name__ == '__main__':

    test_case_0()

    test_case_1()

    print('Test passed')






# coding:utf-8
'''
Dandelion functional pool
Created   :   2, 27, 2018
Revised   :   5, 24, 2018  add `align_crop` which is the same with Lasagne's autocrop()
              8,  3, 2018  move `channel_shuffle()` from `model.shufflenet.py` into `functional.py`
              10, 4, 2018  feature improvement in `spatial_pyramid_pooling` (quanqiu.nick@foxmail.com)
All rights reserved
'''
__author__ = 'dawei.leng'
import theano
floatX = theano.config.floatX
from theano import tensor
from theano.tensor.signal import pool
from .util import as_tuple
from theano.tensor.nnet.neighbours import images2neibs

def pool_1d(x, ws=2, ignore_border=True, stride=None, pad=0, mode='max', axis=-1):
    """
    Pooling 1 dimension along the given axis, support for any dimensional input
    :param x: input to be pooled, can have any dimension
    :param ws:
    :param ignore_border:
    :param stride: if None, same with ws
    :param pad:
    :param mode:
    :param axis: default the last dimension
    :return:
    """
    if stride is not None:
        stride = (stride, 1)
    ndim = x.ndim
    if axis < 0:
        axis += ndim
    pattern = []
    pattern_reverse = list(range(ndim))
    for i in range(ndim):
        if i != axis:
            pattern.append(i)
            if i != len(pattern)-1:
                pattern_reverse[i] = len(pattern)-1
    pattern.extend([axis, 'x'])
    pattern_reverse[axis] = ndim-1
    x = x.dimshuffle(pattern)
    x = pool.pool_2d(x, ws=(ws, 1), ignore_border=ignore_border, stride=stride, pad=(pad,0), mode=mode)[..., 0]
    x = x.dimshuffle(pattern_reverse)
    return x

def pool_2d(x, ws=(2,2), ignore_border=True, stride=None, pad=(0, 0), mode='max'):
    """
    Pooling 2 dimension along the last 2 dimensions of input, support for any dimensional input with ndim>=2
    :param x:
    :param ws:
    :param ignore_border:
    :param stride:
    :param pad:
    :param mode:
    :return:
    """
    return pool.pool_2d(x, ws=ws, ignore_border=ignore_border, stride=stride, pad=pad, mode=mode)

def pool_3d(x, ws=(2,2,2), ignore_border=True, stride=None, pad=(0, 0, 0), mode='max'):
    """
    Pooling 3 dimension along the last 3 dimensions of input, support for any dimensional input with ndim>=3.
    :param x:
    :param ws:
    :param ignore_border:
    :param stride:
    :param pad:
    :param mode:
    :return:
    """
    return pool.pool_3d(x, ws=ws, ignore_border=ignore_border, stride=stride, pad=pad, mode=mode)

def align_crop(tensor_list, cropping):
    """
    align and crop the input tensors by specified rules. Basically the same with Lasagne's autocrop().

    Cropping takes a sequence of tensor_list and crops them per-axis in order to
    ensure that their sizes are consistent so that they can be combined
    in an element-wise fashion. If cropping is enabled for a specific axis,
    the minimum size in that axis of all tensor_list is computed, and all
    tensor_list are cropped to that size.

    The per-axis cropping modes are:

    `None`: this axis is not cropped, tensor_list are unchanged in this axis

    `'lower'`: tensor_list are cropped choosing the lower portion in this axis
    (`a[:crop_size, ...]`)

    `'upper'`: tensor_list are cropped choosing the upper portion in this axis
    (`a[-crop_size:, ...]`)

    `'center'`: tensor_list are cropped choosing the central portion in this axis
    (``a[offset:offset+crop_size, ...]`` where
    ``offset = (a.shape[0]-crop_size)//2)``

    Parameters
    ----------
    tensor_list : list of Theano expressions
        The input arrays in the form of a list of Theano expressions

    cropping : list of cropping modes
        Cropping modes, one for each axis. If length of `cropping` is less
        than the number of axes in the tensor_list, it is padded with `None`.
        If `cropping` is None, `input` is returned as is.

    Returns
    -------
    list of Theano expressions
        each expression is the cropped version of the corresponding input

    Example
    -------
    For example, given three tensor_list:

    >>> import numpy
    >>> import theano

    >>> a = numpy.random.random((1, 2, 3, 4))
    >>> b = numpy.random.random((5, 4, 4, 2))
    >>> c = numpy.random.random((7, 1, 8, 9))

    Cropping mode for each axis:

    >>> cropping = [None, 'lower', 'center', 'upper']

    Crop (note that the input arrays are converted to Theano vars first,
    and that the results are converted back from Theano expressions to
    numpy arrays by calling `eval()`)
    >>> xa, xb, xc = autocrop([theano.shared(a), \
                               theano.shared(b), \
                               theano.shared(c)], cropping)
    >>> xa, xb, xc = xa.eval(), xb.eval(), xc.eval()

    They will be left as is in axis 0 and cropped in the other three,
    choosing the lower, center and upper portions:

    Axis 0: choose all, axis 1: lower 1 element,
    axis 2: central 3 (all) and axis 3: upper 2
    >>> (xa == a[:, :1, :3, -2:]).all()
    True

    Axis 0: choose all, axis 1: lower 1 element,
    axis 2: central 3 starting at 0 and axis 3: upper 2 (all)
    >>> (xb == b[:, :1, :3, -2:]).all()
    True

    Axis 0: all, axis 1: lower 1 element (all),
    axis 2: central 3 starting at 2 and axis 3: upper 2
    >>> (xc == c[:, :1, 2:5:, -2:]).all()
    True
    """
    if cropping is None:
        # No cropping in any dimension
        return tensor_list
    else:
        # Get the number of dimensions
        ndim = tensor_list[0].ndim
        # Check for consistent number of dimensions
        if not all(input.ndim == ndim for input in tensor_list):
            raise ValueError("Not all tensor_list are of the same "
                             "dimensionality. Got {0} tensor_list of "
                             "dimensionalities {1}.".format(
                                len(tensor_list),
                                [input.ndim for input in tensor_list]))
        # Get the shape of each input, where each shape will be a Theano
        # expression
        shapes = [input.shape for input in tensor_list]
        # Convert the shapes to a matrix expression
        shapes_tensor = tensor.as_tensor_variable(shapes)
        # Min along axis 0 to get the minimum size in each dimension
        min_shape = tensor.min(shapes_tensor, axis=0)

        # Nested list of slices; each list in `slices` corresponds to
        # an input and contains a slice for each dimension
        slices_by_input = [[] for i in range(len(tensor_list))]

        # If there are more dimensions than cropping entries, pad the cropping rules with None
        cropping = list(cropping)
        if ndim > len(cropping):
            cropping = list(cropping) + [None] * (ndim - len(cropping))

        # For each dimension
        for dim, cr in enumerate(cropping):
            if cr is None:
                # Don't crop this dimension
                slice_all = slice(None)
                for slices in slices_by_input:
                    slices.append(slice_all)
            else:
                # We crop all tensor_list in the dimension `dim` so that they
                # are the minimum found in this dimension from all tensor_list
                sz = min_shape[dim]
                if cr == 'lower':
                    # Choose the first `sz` elements
                    slc_lower = slice(None, sz)
                    for slices in slices_by_input:
                        slices.append(slc_lower)
                elif cr == 'upper':
                    # Choose the last `sz` elements
                    slc_upper = slice(-sz, None)
                    for slices in slices_by_input:
                        slices.append(slc_upper)
                elif cr == 'center':
                    # Choose `sz` elements from the center
                    for sh, slices in zip(shapes, slices_by_input):
                        offset = (sh[dim] - sz) // 2
                        slices.append(slice(offset, offset+sz))
                else:
                    raise ValueError(
                        'Unknown crop mode \'{0}\''.format(cr))

        return [input[slices] for input, slices in
                zip(tensor_list, slices_by_input)]


def spp_pad_zero(section, pyramid_dim, axis=2, dtype='float32'):
    """
    For spatial_pyramid_pooling method.
    """
    delta = pyramid_dim - section.shape[axis]
    if 2 == axis:
        p = tensor.zeros(shape=(section.shape[0], section.shape[1], delta, section.shape[3]), dtype=dtype)
    elif 3 == axis:
        p = tensor.zeros(shape=(section.shape[0], section.shape[1], section.shape[2], delta), dtype=dtype)
    return tensor.concatenate([section, p], axis=axis)


def spatial_pyramid_pooling(x, pyramid_dims=(6, 4, 2, 1), mode='max', implementation='fast'):
    """
    Spatial pyramid pooling. Refer to: He, Kaiming et al (2015), Spatial Pyramid Pooling in Deep Convolutional Networks
    for Visual Recognition. http://arxiv.org/pdf/1406.4729.pdf and Lasagne's SpatialPyramidPoolingLayer implementation.
    This function will generate spatially fix-sized output no matter the spatial size of input, useful when CNN+FC used for
    image classification or detection.
    :param x: 4D tensor, (B, C, H, W)
    :param pyramid_dims: list of pyramid dims,
    :param mode:
    :param implementation : string
        'fast' | 'fast_ls' | 'stretch'.
        The 'fast' implementation is fast and pad zero when input size is too small.
        The 'fast_ls' implementation is same as Lasagne fast implementation. The size of the input map MUST be larger
            than the output map size.
        The 'stretch' implementation is slower. The implementation will get same feature at some position just like
            nearest neighbor interpolation when the input size is less than the output size.
    :return:
    """
    from theano.ifelse import ifelse
    input_size = x.shape[2:]
    section_list = []

    if 'fast' == implementation:
        for pyramid_dim in pyramid_dims:
            win_size = tuple((i + pyramid_dim - 1) // pyramid_dim for i in input_size)
            str_size = tuple((i + pyramid_dim - 1) // pyramid_dim for i in input_size)
            section = pool.pool_2d(x,
                           ws=win_size,
                           stride=str_size,
                           mode=mode,
                           pad=None,
                           ignore_border=False)

            section = ifelse(tensor.lt(section.shape[2], pyramid_dim),
                             spp_pad_zero(section, pyramid_dim, axis=2, dtype=x.dtype), section)
            section = ifelse(tensor.lt(section.shape[3], pyramid_dim),
                             spp_pad_zero(section, pyramid_dim, axis=3, dtype=x.dtype), section)

            section = section.flatten(3)
            section_list.append(section)

    elif 'fast_ls' == implementation:
        for pyramid_dim in pyramid_dims:
            win_size = tuple((i + pyramid_dim - 1) // pyramid_dim for i in input_size)
            str_size = tuple(i // pyramid_dim for i in input_size)
            section = pool.pool_2d(x,
                           ws=win_size,
                           stride=str_size,
                           mode=mode,
                           pad=None,
                           ignore_border=True)
            section = section.flatten(3)
            section_list.append(section)

    elif 'stretch' == implementation:
        for pyramid_dim in pyramid_dims:
            h, w = input_size
            h = h.astype('float32')
            w = w.astype('float32')
            n = float(pyramid_dim)
            for row in range(pyramid_dim):
                for col in range(pyramid_dim):
                    row = float(row)
                    col = float(col)
                    start_h = tensor.floor(row / n * h).astype('int32')
                    end_h = tensor.ceil((row + 1.0) / n * h).astype('int32')
                    start_w = tensor.floor(col / n * w).astype('int32')
                    end_w = tensor.ceil((col + 1.0) / n * w).astype('int32')

                    pooling_region = x[:, :, start_h:end_h, start_w:end_w]

                    win_size = (end_h-start_h, end_w-start_w)
                    str_size = (end_h-start_h, end_w-start_w)
                    section = pool.pool_2d(pooling_region,
                                           ws=win_size,
                                           stride=str_size,
                                           mode=mode,
                                           pad=None,
                                           ignore_border=False)
                    section = section.flatten(3)
                    section_list.append(section)

    return tensor.concatenate(section_list, axis=2)


def upsample_2d(x, ratio, mode='repeat'):
    """
    Upsample 2 dimension along the last 2 dimensions of input, support for any dimensional input with ndim>=2.
    :param x:
    :param ratio: must be integer or tuple of integers >=1
    :param mode: 'repeat' | 'dilate'. Repeat element values or upsample leaving zeroes between upsampled elements
    :return:
    """
    a, b = as_tuple(ratio, 2)
    upscaled = x
    if mode == 'repeat':
        if b > 1:
            upscaled = tensor.extra_ops.repeat(upscaled, b, axis=-1)
        if a > 1:
            upscaled = tensor.extra_ops.repeat(upscaled, a, axis=-2)
    elif mode == 'dilate':
        if b > 1 or a > 1:
            output_shape = list(x.shape)  # copy / convert to mutable list
            if output_shape[-2] is not None:
                output_shape[-2] *= a
            if output_shape[-1] is not None:
                output_shape[-1] *= b
            upscaled = tensor.zeros(shape=tuple(output_shape), dtype=x.dtype)
            upscaled = tensor.set_subtensor(upscaled[:, :, ::a, ::b], x)
    else:
        raise ValueError('mode = %s not supported' % mode)
    return upscaled

def upsample_2d_bilinear(x, ratio=None, frac_ratio=None, use_1D_kernel=True):
    """
    Upsample 2D with bilinear interpolation. Support for fractional ratio, and only apply for 4D tensor.
    :param x: 4D tensor.
    :param ratio: int or tuple of int >=1. You can only specify either `ratio` or `frac_ratio`, not both.
    :param frac_ratio: None, tuple of int or tuple of tuples of int. A fractional upsampling scale is described by
                       (numerator, denominator).
    :param use_1D_kernel: only for speed matter
    :return:
    """
    return tensor.nnet.abstract_conv.bilinear_upsampling(x, ratio=ratio, frac_ratio=frac_ratio, use_1D_kernel=use_1D_kernel)

def channel_shuffle(x, group_num):
    """
    Pseudo shuffle channel by dimshuffle & reshape, first introduced in [ShuffleNet](https://arxiv.org/abs/1610.02357)
    :param x: (B, C, H, W)
    :param group_num: int scalar, C must be divisible by group_num
    :return:
    """
    if group_num == 1:
        return x
    B, C, H, W = x.shape
    x = tensor.reshape(x, (B, group_num, C//group_num, H, W))
    x = x.dimshuffle(0, 2, 1, 3, 4)
    x = tensor.reshape(x, (B, -1, H, W))
    return x

#todo: due to theano's images2neibs(), this tensor function does not support gradient computation [7-30-2018]
#todo: this API won't be exposed until we find another way to implement `im2col` with gradient support
#todo: if this API is deleted, test/test_im2col.py should be deleted as well
def _im2col(x, nb_size=(3,3), border_mode='half', merge_channel=False):
    """
    Function im2col allows to apply a sliding window operation to a 4D tensor, based on theano's images2neibs() and corresponding to Caffe's im2col()
    :todo: support any nb_step
    :param x: must be 4D tensor with shape (B, C, H, W)
    :param nb_size: neighbour size (h, w), i.e., the sliding window size, 2D vector. Only odd shapes are supported.
    :param border_mode: 'half' | 'wrap_centered' : on image border, whether pad by 0 or wrap image
    :param merge_channel: if False, return with shape (B, C, H, W, h*w), otherwise return with shape (B, H, W, C*h*w), corresponding to Caffe's im2col()
    :return:
    """
    nb_step = (1, 1)
    B, C, H, W = x.shape
    x = images2neibs(x, neib_shape=nb_size, neib_step=nb_step, mode=border_mode) # (B*C*H*W, h*w)
    x = tensor.reshape(x, (B, C, H, W, -1))  # (B, C, H, W, h*w)
    if merge_channel:
        x = x.dimshuffle(0, 2, 3, 1, 4)  # (B, H, W, C, h*w)
        x = tensor.reshape(x, (B, H, W, -1))  # (B, H, W, C*h*w)
    return x



if __name__ == '__main__':
    INFO = ['Dandelion framework: module pool\n',
            'Author: David Leon\n',
            'All rights reserved\n']
    print(*INFO)






# coding:utf-8
'''
Reference implementation of [CTPN](https://arxiv.org/abs/1609.03605) for text line detection
Created   :   7, 26, 2018
Revised   :   8,  2, 2018  add alternative implementation to bypass the absense of `im2col` in Theano.
              9,  3, 2018  modified: change output bbox's shape to (B, H, W, k, n) (from (B, H, W, k*n)).
All rights reserved
'''
# ------------------------------------------------------------------------------------------------
__author__ = 'dawei.leng'

import theano.tensor as tensor
from ..module import *
from ..functional import *
from ..activation import *


class model_CTPN(Module):
    """
    Reference implementation of [CTPN](https://arxiv.org/abs/1609.03605) for text line detection
    """

    def __init__(self, k=10, do_side_refinement_regress=False,
                 batchnorm_mode=1, channel=3, im_height=None, im_width=None,
                 kernel_size=3, border_mode=(1, 1), VGG_flip_filters=False,
                 im2col=None):
        """

        :param k: anchor box number
        :param do_side_refinement_regress: whether implement side refinement regression
        :param batchnorm_mode: 1|0, whether insert batch normalization into the end of each convolution stage of VGG-16 net
        :param channel: input channel number
        :param im_height: input image height, optional
        :param im_width:  input image width, optional
        :param kernel_size: convolution kernel size of VGG-16 net
        :param border_mode: border mode of VGG-16 net
        :param VGG_flip_filters: whether flip convolution kernels for VGG-16 net
        :param im2col: function corresponding to Caffe's `im2col()`. If None, the CTPN implementation will not strictly follow the original paper.
        """
        super().__init__()
        self.batchnorm_mode = batchnorm_mode
        self.k = k
        self.im2col = im2col
        #--- encoding part, VGG16, 13 conv layers ---#
        self.conv1_1 = Conv2D(in_channels=channel, out_channels=64, kernel_size=kernel_size, pad=border_mode, input_shape=(im_height, im_width), flip_filters=VGG_flip_filters)
        self.conv1_2 = Conv2D(in_channels=64,  out_channels=64,  kernel_size=kernel_size, pad=border_mode, flip_filters=VGG_flip_filters)
        self.conv2_1 = Conv2D(in_channels=64,  out_channels=128, kernel_size=kernel_size, pad=border_mode, flip_filters=VGG_flip_filters)
        self.conv2_2 = Conv2D(in_channels=128, out_channels=128, kernel_size=kernel_size, pad=border_mode, flip_filters=VGG_flip_filters)
        self.conv3_1 = Conv2D(in_channels=128, out_channels=256, kernel_size=kernel_size, pad=border_mode, flip_filters=VGG_flip_filters)
        self.conv3_2 = Conv2D(in_channels=256, out_channels=256, kernel_size=kernel_size, pad=border_mode, flip_filters=VGG_flip_filters)
        self.conv3_3 = Conv2D(in_channels=256, out_channels=256, kernel_size=kernel_size, pad=border_mode, flip_filters=VGG_flip_filters)
        self.conv4_1 = Conv2D(in_channels=256, out_channels=512, kernel_size=kernel_size, pad=border_mode, flip_filters=VGG_flip_filters)
        self.conv4_2 = Conv2D(in_channels=512, out_channels=512, kernel_size=kernel_size, pad=border_mode, flip_filters=VGG_flip_filters)
        self.conv4_3 = Conv2D(in_channels=512, out_channels=512, kernel_size=kernel_size, pad=border_mode, flip_filters=VGG_flip_filters)
        self.conv5_1 = Conv2D(in_channels=512, out_channels=512, kernel_size=kernel_size, pad=border_mode, flip_filters=VGG_flip_filters)
        self.conv5_2 = Conv2D(in_channels=512, out_channels=512, kernel_size=kernel_size, pad=border_mode, flip_filters=VGG_flip_filters)
        self.conv5_3 = Conv2D(in_channels=512, out_channels=512, kernel_size=kernel_size, pad=border_mode, flip_filters=VGG_flip_filters)  # (B, C, H, W)
        if batchnorm_mode == 0:
            pass
        elif batchnorm_mode == 1:
            self.bn1 = BatchNorm(input_shape=(None,  64, None, None))
            self.bn2 = BatchNorm(input_shape=(None, 128, None, None))
            self.bn3 = BatchNorm(input_shape=(None, 256, None, None))
            self.bn4 = BatchNorm(input_shape=(None, 512, None, None))
            self.bn5 = BatchNorm(input_shape=(None, 512, None, None))
        else:
            raise ValueError('batchnorm_mode should = 0 or 1')
        #--- detection part ---#
        if im2col is not None:  # implementation strictly follow the reference paper
            self.lstm_f    = LSTM(input_dims=512*9, hidden_dim=128, grad_clipping=100) # (W, B*H, C)
            self.lstm_b    = LSTM(input_dims=512*9, hidden_dim=128, grad_clipping=100) # (W, B*H, C)
            self.conv_rpn  = Conv2D(in_channels=256, out_channels=512, kernel_size=1, pad='same')  # (B, C, H, W)
            self.conv_cls_score = Conv2D(in_channels=512, out_channels=2 * k, kernel_size=1, pad='same')  # (B, 2*k, H, W)
            if do_side_refinement_regress:
                self.conv_bbox_pred = Conv2D(in_channels=512, out_channels=3 * k, kernel_size=1, pad='same')  # (B, 3*k, H, W), include side-refinement
            else:
                self.conv_bbox_pred = Conv2D(in_channels=512, out_channels=2 * k, kernel_size=1, pad='same')  # (B, 2*k, H, W), no side-refinement

        else: # implementation putting convolution before RNN, doesn't follow the reference paper
            self.conv_rpn = Conv2D(in_channels=512, out_channels=512, kernel_size=(3,3), pad=(1,1))  # (B, C, H, W)
            self.lstm_f   = LSTM(input_dims=512, hidden_dim=128, grad_clipping=100)  # (W, B*H, C)
            self.lstm_b   = LSTM(input_dims=512, hidden_dim=128, grad_clipping=100)  # (W, B*H, C)
            self.conv_cls_score = Conv2D(in_channels=256, out_channels=2*k, kernel_size=1, pad='same')  # (B, 2*k, H, W)
            if do_side_refinement_regress:
                self.conv_bbox_pred = Conv2D(in_channels=256, out_channels=3*k, kernel_size=1, pad='same')  # (B, 3*k, H, W), include side-refinement
            else:
                self.conv_bbox_pred = Conv2D(in_channels=256, out_channels=2*k, kernel_size=1, pad='same')  # (B, 2*k, H, W), no side-refinement

    def forward(self, x):
        """
        :param x: (B, C, H, W)
        :return: cls_score, bbox  # (B, H, W, k, 2), (B, H, W, n*k) n = 2 or 3
        """
        self.work_mode = 'train'

        #--- encoding part, VGG16, 13 conv layers ---#
        x = relu(self.conv1_1.forward(x))   # (B, 64, 224, 224)
        x = relu(self.conv1_2.forward(x))
        if self.batchnorm_mode == 1:
            x = self.bn1.forward(x)
        x = pool_2d(x, ws=(2, 2))           # (B, 64, 112, 112)
        x = relu(self.conv2_1.forward(x))
        x = relu(self.conv2_2.forward(x))
        if self.batchnorm_mode == 1:
            x = self.bn2.forward(x)
        x = pool_2d(x, ws=(2, 2))           # (B, 128, 56, 56)
        x = relu(self.conv3_1.forward(x))
        x = relu(self.conv3_2.forward(x))
        x = relu(self.conv3_3.forward(x))
        if self.batchnorm_mode == 1:
            x = self.bn3.forward(x)
        x = pool_2d(x, ws=(2, 2))           # (B, 256, 28, 28)
        x = relu(self.conv4_1.forward(x))
        x = relu(self.conv4_2.forward(x))
        x = relu(self.conv4_3.forward(x))
        if self.batchnorm_mode == 1:
            x = self.bn4.forward(x)
        x = pool_2d(x, ws=(2, 2))           # (B, 512, 14, 14)
        x = relu(self.conv5_1.forward(x))
        x = relu(self.conv5_2.forward(x))
        x = relu(self.conv5_3.forward(x))
        if self.batchnorm_mode == 1:
            x = self.bn5.forward(x)
        #--- detection part ---#
        if self.im2col is not None:
            x = self.im2col(x, nb_size=(3,3), merge_channel=True)   # (B, H, W, C*h*w)
            B, H, W, Chw = x.shape
            x = x.dimshuffle((2, 0, 1, 3))
            x = x.reshape((W, -1, Chw))   # (W, B*H, C*h*w)
        else:
            x = relu(self.conv_rpn.forward(x))
            B, C, H, W = x.shape
            x = x.dimshuffle((3, 0, 2, 1)) # (W, B, H, C)
            x = x.reshape((W, -1, C))      # (W, B*H, C)

        x_f = self.lstm_f.forward(x)
        x_b = self.lstm_b.forward(x, backward=True)
        x = tensor.concatenate([x_f, x_b], axis=2)  # (W, B*H, 256)
        x = x.reshape((W, B, H, -1))
        x = x.dimshuffle((1, 3, 2, 0)) # (B, 256, H, W)
        if self.im2col is not None:
            x = relu(self.conv_rpn.forward(x))   # (B, 512, H, W)

        cls_score = self.conv_cls_score.forward(x)  # (B, 2*k, H, W)
        cls_score = cls_score.reshape((B, 2, self.k, H, W)) # (B, 2, k, H, W)
        cls_score = softmax(cls_score.dimshuffle((0, 3, 4, 2, 1)))  # (B, H, W, k, 2)
        bbox = self.conv_bbox_pred.forward(x) # (B, 3*k, H, W), no activation applied
        bbox = bbox.dimshuffle((0, 2, 3, 1, 'x'))  # (B, H, W, 3*k, 1)
        bbox = bbox.reshape((B, H, W, self.k, -1)) # (B, H, W, k, n)

        return cls_score, bbox  # (B, H, W, k, 2), (B, H, W, k, n) n = 2 or 3

    def predict(self, x):
        self.work_mode = 'inference'

        #--- encoding part, VGG16, 13 conv layers ---#
        x = relu(self.conv1_1.predict(x))   # (B, 64, 224, 224)
        x = relu(self.conv1_2.predict(x))
        if self.batchnorm_mode == 1:
            x = self.bn1.predict(x)
        x = pool_2d(x, ws=(2, 2))           # (B, 64, 112, 112)
        x = relu(self.conv2_1.predict(x))
        x = relu(self.conv2_2.predict(x))
        if self.batchnorm_mode == 1:
            x = self.bn2.predict(x)
        x = pool_2d(x, ws=(2, 2))           # (B, 128, 56, 56)
        x = relu(self.conv3_1.predict(x))
        x = relu(self.conv3_2.predict(x))
        x = relu(self.conv3_3.predict(x))
        if self.batchnorm_mode == 1:
            x = self.bn3.predict(x)
        x = pool_2d(x, ws=(2, 2))           # (B, 256, 28, 28)
        x = relu(self.conv4_1.predict(x))
        x = relu(self.conv4_2.predict(x))
        x = relu(self.conv4_3.predict(x))
        if self.batchnorm_mode == 1:
            x = self.bn4.predict(x)
        x = pool_2d(x, ws=(2, 2))           # (B, 512, 14, 14)
        x = relu(self.conv5_1.predict(x))
        x = relu(self.conv5_2.predict(x))
        x = relu(self.conv5_3.predict(x))
        if self.batchnorm_mode == 1:
            x = self.bn5.predict(x)
        #--- detection part ---#
        if self.im2col is not None:
            x = self.im2col(x, nb_size=(3,3), merge_channel=True)   # (B, H, W, C*h*w)
            B, H, W, Chw = x.shape
            x = x.dimshuffle((2, 0, 1, 3))
            x = x.reshape((W, -1, Chw))   # (W, B*H, C*h*w)
        else:
            x = relu(self.conv_rpn.predict(x))
            B, C, H, W = x.shape
            x = x.dimshuffle((3, 0, 2, 1)) # (W, B, H, C)
            x = x.reshape((W, -1, C))      # (W, B*H, C)

        x_f = self.lstm_f.predict(x)
        x_b = self.lstm_b.predict(x, backward=True)
        x = tensor.concatenate([x_f, x_b], axis=2)  # (W, B*H, 256)
        x = x.reshape((W, B, H, -1))
        x = x.dimshuffle((1, 3, 2, 0)) # (B, 256, H, W)
        if self.im2col is not None:
            x = relu(self.conv_rpn.predict(x))   # (B, 512, H, W)

        cls_score = self.conv_cls_score.predict(x)  # (B, 2*k, H, W)
        cls_score = cls_score.reshape((B, 2, self.k, H, W)) # (B, 2, k, H, W)
        cls_score = softmax(cls_score.dimshuffle((0, 3, 4, 2, 1)))  # (B, H, W, k, 2)
        bbox = self.conv_bbox_pred.predict(x) # (B, 3*k, H, W), no activation applied
        bbox = bbox.dimshuffle((0, 2, 3, 1, 'x'))  # (B, H, W, 3*k, 1)
        bbox = bbox.reshape((B, H, W, self.k, -1)) # (B, H, W, k, n)

        return cls_score, bbox  # (B, H, W, k, 2), (B, H, W, k, n) n = 2 or 3


# coding:utf-8
# Test for feature pyramid net, partial.
# Created   :   7,  6, 2018
# Revised   :   7,  6, 2018
# All rights reserved
#------------------------------------------------------------------------------------------------
__author__ = 'dawei.leng'
import os, sys
os.environ['THEANO_FLAGS'] = "floatX=float32, mode=FAST_RUN, warn_float64='raise'"

import theano
from theano import tensor
from dandelion.module import *
from dandelion.activation import *
from dandelion.model.feature_pyramid_net import model_FPN

import dandelion
dandelion_path = os.path.split(dandelion.__file__)[0]
print('dandelion path = %s\n' % dandelion_path)

def test_case_0():
    model = model_FPN(input_channel=1, batchnorm_mode=2, base_n_filters=32)
    x = tensor.ftensor4('x')
    y = model.forward(x)
    print('compiling fn...')
    fn = theano.function([x], y, no_default_updates=False)
    print('run fn...')
    input = np.random.rand(2, 1, 64, 63).astype(np.float32)
    output = fn(input)
    for r in output:
        print(r.shape)
    # print(output.shape)
    print(output)

if __name__ == '__main__':

    test_case_0()

    print('Test passed')






# coding:utf-8
# Test ChainCRF() class with anago's implementation
# Created   :   2, 12, 2018
# Revised   :   2, 12, 2018
# All rights reserved
#------------------------------------------------------------------------------------------------
__author__ = 'dawei.leng'
import os, sys
os.environ['THEANO_FLAGS'] = "floatX=float32, mode=FAST_RUN, warn_float64='raise'"

from keras import backend as K
from keras import initializers, regularizers, constraints
from keras.engine import Layer, InputSpec
import theano, numpy as np
import theano.tensor as tensor

import dandelion
dandelion_path = os.path.split(dandelion.__file__)[0]
print('dandelion path = %s\n' % dandelion_path)

def log_sum_exp(x, axis=None, keepdims=False):
    """
    Stable log of a sum of exponentials
    """
    x_max = tensor.max(x, axis=axis, keepdims=True)
    z = tensor.log(tensor.sum(tensor.exp(x - x_max), axis=axis, keepdims=True)) + x_max
    return z.sum(axis=axis, keepdims=keepdims)

def path_energy(y, x, U, b_start=None, b_end=None, mask=None):
    """Calculates the energy of a tag path y for a given input x (with mask),
    transition energies U and boundary energies b_start, b_end."""
    x = add_boundary_energy(x, b_start, b_end, mask)
    return path_energy0(y, x, U, mask)


def path_energy0(y, x, U, mask=None):
    """Path energy without boundary potential handling."""
    n_classes = K.shape(x)[2]                   # x.shape = (B, T, N)
    y_one_hot = K.one_hot(y, n_classes)         # convert integer 'y' to one-hot encoded 'y': (B, T) -> (B, T, N)

    # Tag path energy
    energy = K.sum(x * y_one_hot, 2)    # (B, T, N) -> (B, T)
    energy = K.sum(energy, 1)           # (B, T) -> (B,)

    # Transition energy
    y_t = y[:, :-1]                     # y_t, (B, T-1)
    y_tp1 = y[:, 1:]                    # y_(t+1), (B, T-1)
    U_flat = K.reshape(U, [-1])         # (N, N) -> (N*N,)
    # Convert 2-dim indices (y_t, y_tp1) of U to 1-dim indices of U_flat:
    flat_indices = y_t * n_classes + y_tp1
    U_y_t_tp1 = K.gather(U_flat, flat_indices)

    if mask is not None:
        mask = K.cast(mask, K.floatx())
        y_t_mask = mask[:, :-1]
        y_tp1_mask = mask[:, 1:]
        U_y_t_tp1 *= y_t_mask * y_tp1_mask

    energy += K.sum(U_y_t_tp1, axis=1)

    return energy  #(B,)


def sparse_chain_crf_loss(y, x, U, b_start=None, b_end=None, mask=None):
    """Given the true sparsely encoded tag sequence y, input x (with mask),
    transition energies U, boundary energies b_start and b_end, it computes
    the loss function of a Linear Chain Conditional Random Field:
    loss(y, x) = NLL(P(y|x)), where P(y|x) = exp(E(y, x)) / Z.
    So, loss(y, x) = - E(y, x) + log(Z)
    Here, E(y, x) is the tag path energy, and Z is the normalization constant.
    The values log(Z) is also called free energy.
    """
    x = add_boundary_energy(x, b_start, b_end, mask)
    energy = path_energy0(y, x, U, mask)
    energy -= free_energy0(x, U, mask)
    return K.expand_dims(-energy, -1)


def chain_crf_loss(y, x, U, b_start=None, b_end=None, mask=None):
    """Variant of sparse_chain_crf_loss but with one-hot encoded tags y."""
    y_sparse = K.argmax(y, -1)
    y_sparse = K.cast(y_sparse, 'int32')
    return sparse_chain_crf_loss(y_sparse, x, U, b_start, b_end, mask)


def add_boundary_energy(x, b_start=None, b_end=None, mask=None):
    """Given the observations x, it adds the start boundary energy b_start (resp.
    end boundary energy b_end on the start (resp. end) elements and multiplies
    the mask."""
    if mask is None:
        if b_start is not None:
            x = K.concatenate([x[:, :1, :] + b_start, x[:, 1:, :]], axis=1)  # dim_1 is T
        if b_end is not None:
            x = K.concatenate([x[:, :-1, :], x[:, -1:, :] + b_end], axis=1)
    else:
        mask = K.cast(mask, K.floatx())
        mask = K.expand_dims(mask, 2)
        x *= mask
        if b_start is not None:
            mask_r = K.concatenate([K.zeros_like(mask[:, :1]), mask[:, :-1]], axis=1)
            start_mask = K.cast(K.greater(mask, mask_r), K.floatx())
            x = x + start_mask * b_start
        if b_end is not None:
            mask_l = K.concatenate([mask[:, 1:], K.zeros_like(mask[:, -1:])], axis=1)
            end_mask = K.cast(K.greater(mask, mask_l), K.floatx())
            x = x + end_mask * b_end
    return x


def viterbi_decode(x, U, b_start=None, b_end=None, mask=None):
    """Computes the best tag sequence y for a given input x, i.e. the one that
    maximizes the value of path_energy."""
    x = add_boundary_energy(x, b_start, b_end, mask)

    alpha_0 = x[:, 0, :]   # (B, N)
    gamma_0 = K.zeros_like(alpha_0)
    initial_states = [gamma_0, alpha_0]
    # the following ``` lambda B: [K.cast(K.argmax(B, axis=1), K.floatx()), K.max(B, axis=1)], ``` means [idx_max, value_max]
    _, gamma = _forward(x,
                        lambda B: [K.cast(K.argmax(B, axis=1), K.floatx()), K.max(B, axis=1)],
                        initial_states,
                        U,
                        mask)
    # gamma: (B, T, N)
    y = _backward(gamma, mask)
    return y


def free_energy(x, U, b_start=None, b_end=None, mask=None):
    """Computes efficiently the sum of all path energies for input x, when
    runs over all possible tag sequences."""
    x = add_boundary_energy(x, b_start, b_end, mask)
    return free_energy0(x, U, mask)


def free_energy0(x, U, mask=None):
    """Free energy without boundary potential handling.
    x: (B, T, N)
    U: (N, N)
    """
    initial_states = [x[:, 0, :]]         # [B, N]
    last_alpha, _ = _forward(x,
                             lambda B: [K.logsumexp(B, axis=1)],
                             initial_states,
                             U,
                             mask)
    return last_alpha[:, 0]


def _forward(x, reduce_step, initial_states, U, mask=None):
    """Forward recurrence of the linear chain crf."""

    def _forward_step(energy_matrix_t, states):  # (B, N, N), [(N,), (B, N)]
        alpha_tm1 = states[-1]
        new_states = reduce_step(K.expand_dims(alpha_tm1, 2) + energy_matrix_t)
        return new_states[0], new_states

    U_shared = K.expand_dims(K.expand_dims(U, 0), 0)   # (N, N) -> (1, 1, N, N)

    if mask is not None:
        mask = K.cast(mask, K.floatx())
        mask_U = K.expand_dims(K.expand_dims(mask[:, :-1] * mask[:, 1:], 2), 3)
        U_shared = U_shared * mask_U

    inputs = K.expand_dims(x[:, 1:, :], 2) + U_shared     # (B, T-1, 1, N) + (1, 1, N, N) -> (B, T-1, N, N)
    inputs = K.concatenate([inputs, K.zeros_like(inputs[:, -1:, :, :])], axis=1) # (B, T-1, N, N) -> (B, T, N, N)

    last, values, _ = K.rnn(_forward_step, inputs, initial_states)
    return last, values


def batch_gather(reference, indices):
    """

    :param reference: (B, N)
    :param indices: (B,)
    :return:
    """
    ref_shape = K.shape(reference)
    batch_size = ref_shape[0]
    n_classes = ref_shape[1]
    flat_indices = K.arange(0, batch_size) * n_classes + K.flatten(indices)
    return K.gather(K.flatten(reference), flat_indices)


def _backward(gamma, mask):
    """Backward recurrence of the linear chain crf."""
    gamma = K.cast(gamma, 'int32')   # (B, T, N)

    def _backward_step(gamma_t, states):
        # print('len(states)=', len(states))
        # print(type(states))
        # y_tm1 = K.squeeze(states[0], 0)
        y_tm1 = states[0]
        y_t = batch_gather(gamma_t, y_tm1)
        # return y_t, [K.expand_dims(y_t, 0)]
        # return K.expand_dims(y_t, 0), [K.expand_dims(y_t, 0)]
        return y_t, [y_t]

    # initial_states = [K.expand_dims(K.zeros_like(gamma[:, 0, 0]), 0)]  # (1, B)
    initial_states = [K.zeros_like(gamma[:, 0, 0])]  # (1, B)
    _, y_rev, _ = K.rnn(_backward_step,
                        gamma,
                        initial_states,
                        go_backwards=True)
    y = K.reverse(y_rev, 1)

    if mask is not None:
        mask = K.cast(mask, dtype='int32')
        # mask output
        y *= mask
        # set masked values to -1
        y += -(1 - mask)
    return y

#--------- ChainCRF() -------------------#
def CRF_forward(observations, transitions):
    """

    :param observations: (B, T, N)
    :param transitions:  (N, N)
    :param viterbi:
    :param return_alpha:
    :param return_best_sequence:
    :return:
    """
    U = transitions.dimshuffle('x', 'x', 0, 1)  # (N, N) -> (1, 1, N, N)
    initial = observations[:, 0, :] # (B, N)
    tensor.unbroadcast(initial, 0, 1)
    x = observations[:, 1:, :]  # (B, T-1, N)
    x = x.dimshuffle(0, 1, 'x', 2) #(B, T-1, N) -> (B, T-1, 1, N)
    x = x + U
    # x = tensor.concatenate([x, tensor.zeros_like(x[:, -1:, :, :])], axis=1)  # (B, T-1, N, N) -> (B, T, N, N)
    x = x.dimshuffle(1, 0, 2, 3) # (T, B, N, N)

    def recurrence(energy_matrix_t, states):
        """
        :param energy_matrix_t: (B, N, N)

        :return:
        """
        alpha_tm1 = states # (B,N)
        alpha_tm1 = alpha_tm1.dimshuffle(0, 1, 'x') # (B, N, 1)
        x = alpha_tm1 + energy_matrix_t  # (B, N, N)
        new_states = log_sum_exp(x, axis=1)  # (B, N)
        return new_states

    # alpha: (T, B, N)
    alpha, _ = theano.scan(
        fn=recurrence,
        sequences=x,
        outputs_info=[initial]
    )
    # return alpha[-1, :, :]  # (B,)
    return alpha.dimshuffle(1, 0, 2)

def CRF_decode(observations, transitions):
    """

    :param observations: (B, T, N)
    :param transitions:  (N, N)
    :param viterbi:
    :param return_alpha:
    :param return_best_sequence:
    :return:
    """
    alpha_0 = observations[:, 0, :] # (B, N)
    gamma_0 = tensor.zeros_like(alpha_0, dtype='int64')

    U = transitions.dimshuffle('x', 'x', 0, 1)   # (N, N) -> (1, 1, N, N)
    x = observations[:, 1:, :]  # (B, T-1, N)
    x = x.dimshuffle(0, 1, 'x', 2) #(B, T-1, N) -> (B, T-1, 1, N)
    x = x + U    # (B, T-1 N, N)
    x = tensor.concatenate([x, tensor.zeros_like(x[:, -1:, :, :])], axis=1)  # (B, T-1, N, N) -> (B, T, N, N)
    x = x.dimshuffle(1, 0, 2, 3) # (T, B, N, N)

    def recurrence(energy_matrix_t, index_tm1, score_tm1):
        """
        :param energy_matrix_t: (B, N, N)

        :return:
        """
        score_tm1 = score_tm1.dimshuffle(0, 1, 'x') # (B, N, 1)
        x = score_tm1 + energy_matrix_t  # (B, N, N)
        index = tensor.argmax(x, axis=1) # (B, N)
        score = tensor.max(x, axis=1)    # (B, N)
        return index, score

    # gamma, alpha: (T, B, N)
    result, _ = theano.scan(
        fn=recurrence,
        sequences=x,
        outputs_info=[gamma_0, alpha_0]
    )
    gamma, alpha = result

    def backward_step(gamma_t, y_tm1):
        y_t   = batch_gather(gamma_t, y_tm1)
        return y_t

    T, B, N = gamma.shape
    initial = tensor.zeros(shape=(B,), dtype='int64')
    # y : (T, B)
    y, _ = theano.scan(
        fn=backward_step,
        sequences=gamma,
        outputs_info=[initial],
        go_backwards=True
    )
    y = y.dimshuffle(1, 0)
    y = y[:, ::-1]
    y = tensor.cast(y, 'int32')
    return y  # (B, T)
    # return gamma.dimshuffle(1, 0, 2)

#--------- ChainCRF_Lample() ------------#
def CRF_forward_nobatch(observations, transitions, viterbi=False, return_alpha=False, return_best_sequence=False):
    """
    Takes as input:
        [DV]  (T+2, N+2), (N+2, N+2)
        - observations, sequence of shape (n_steps, n_classes)
        - transitions, sequence of shape (n_classes, n_classes)
    Probabilities must be given in the log space.
    Compute alpha, matrix of size (n_steps, n_classes), such that
    alpha[i, j] represents one of these 2 values:
        - the probability that the real path at node i ends in j
        - the maximum probability of a path finishing in j at node i (Viterbi)
    Returns one of these 2 values:
        - alpha
        - the final probability, which can be:
            - the sum of the probabilities of all paths
            - the probability of the best path (Viterbi)
    """
    assert not return_best_sequence or (viterbi and not return_alpha)

    def recurrence(obs, previous, transitions):
        """

        :param obs: (N+2,)
        :param previous: (N+2,)
        :param transitions: (N+2, N+2)
        :return:
        """
        previous = previous.dimshuffle(0, 'x')  # (N+2,) -> (N+2, 1)
        obs = obs.dimshuffle('x', 0)  # (N+2,) -> (1, N+2)
        if viterbi:
            scores = previous + obs + transitions
            out = scores.max(axis=0)
            if return_best_sequence:
                out2 = scores.argmax(axis=0)
                return out, out2
            else:
                return out
        else:
            return log_sum_exp(previous + obs + transitions, axis=0)

    initial = observations[0]  # = b_s = [[small] * T, 0, small]
    alpha, _ = theano.scan(
        fn=recurrence,
        outputs_info=(initial, None) if return_best_sequence else initial,
        sequences=[observations[1:]],
        non_sequences=transitions
    )

    if return_alpha:
        return alpha
    elif return_best_sequence:
        sequence, _ = theano.scan(
            fn=lambda beta_i, previous: beta_i[previous],
            outputs_info=tensor.cast(tensor.argmax(alpha[0][-1]), 'int32'),
            sequences=tensor.cast(alpha[1][::-1], 'int32')
        )
        sequence = tensor.concatenate([sequence[::-1], [tensor.argmax(alpha[0][-1])]])
        return sequence
    else:
        if viterbi:
            return alpha[-1].max(axis=0)
        else:
            return log_sum_exp(alpha[-1],
                               axis=0)  # p(x). Here alpha is equivalent to beta in "CRF as NN Layer", Page10.

def disabled_test_case_0():
    y = tensor.imatrix('y')
    x = tensor.ftensor3('x')
    U = tensor.fmatrix('U')
    x_nobatch = tensor.fmatrix('x_nobatch')

    cost1 = free_energy0(x, U)
    cost2 = CRF_forward(x, U)
    cost3 = CRF_forward_nobatch(x_nobatch, U)
    seq1 = viterbi_decode(x, U)
    seq2 = CRF_decode(x, U)
    print('compiling f1 & f2 ...')
    f1 = theano.function([x, U], cost1)
    f2 = theano.function([x, U], cost2)
    print('compiling f3 ...')
    f3 = theano.function([x, U], seq1)
    print('compiling f4 ...')
    f4 = theano.function([x, U], seq2)
    print('compiling f5 ...')
    f5 = theano.function([x_nobatch, U], cost3)

    B, T, N = 7, 100, 20
    for i in range(1):
        x = np.random.rand(B, T, N).astype(np.float32)
        U = np.random.rand(N, N).astype(np.float32)

        c1 = f1(x, U)
        c2 = f2(x, U)
        s1 = f3(x, U)
        s2 = f4(x, U)
        print(c1)
        print(c2)
        print(c1 == c2)
        if np.all(c1 == c2):
            print('c pass')
        else:
            raise ValueError('c not same!')

        print(s1)
        print(s2)
        print(s1.shape)
        print(s2.shape)
        print(s1 == s2)
        if np.all(s1 == s2):
            print('s pass')
        else:
            raise ValueError('s not same!')

def win_test_case_1():
    """
    This test case passed on Tensorflow 1.8.0 + Win64, but failed on Tensorflow 1.8.0 + Ubuntu 14.04
    :return:
    """
    y = tensor.imatrix('y')
    x = tensor.ftensor3('x')
    U = tensor.fmatrix('U')
    x_nobatch = tensor.fmatrix('x_nobatch')

    cost2 = CRF_forward(x, U)
    cost3 = CRF_forward_nobatch(x_nobatch, U, return_alpha=True)
    seq2 = CRF_decode(x, U)
    seq3 = CRF_forward_nobatch(x_nobatch, U, viterbi=True, return_best_sequence=True)
    print('compiling f2 ...')
    f2 = theano.function([x, U], cost2)
    print('compiling f4 ...')
    f4 = theano.function([x, U], seq2)
    print('compiling f5 ...')
    f5 = theano.function([x_nobatch, U], cost3)
    print('compiling f6 ...')
    f6 = theano.function([x_nobatch, U], seq3)

    B, T, N = 2, 5, 3
    for i in range(10):
        x = np.random.rand(B, T, N).astype(np.float32)
        U = np.random.rand(N, N).astype(np.float32)

        c2 = f2(x, U)
        s2 = f4(x, U)

        c3_list = []
        s3_list = []
        for j in range(B):
            c3_nobatch = f5(x[j, :, :], U)
            s3_nobatch = f6(x[j, :, :], U)
            c3_list.append(np.expand_dims(c3_nobatch, 0))
            s3_list.append(np.expand_dims(s3_nobatch, 0))
        c3 = np.concatenate(c3_list, axis=0)
        s3 = np.concatenate(s3_list, axis=0)

        # if np.all(c2 == c3):
        #     print('c pass')
        # else:
        print(c2)
        print(c3)
        # raise ValueError('c not same!')

        print(s2)
        print(s3)
        if np.all(s2 == s3):
            print('s pass')
        else:

            raise ValueError('s not same!')

if __name__ == '__main__':

    # test_case_0()

    win_test_case_1()

    print('Test passed~')






# coding:utf-8
'''
Reference implementation of [Shuffle-net](https://arxiv.org/abs/1707.01083) and other related models in paper.
Created   :   7,  9, 2018
Revised   :   7,  9, 2018
All rights reserved
'''
# ------------------------------------------------------------------------------------------------
__author__ = 'dawei.leng'

import theano.tensor as tensor
from ..module import *
from ..functional import *
from ..activation import *


class DSConv2D(Module):
    """
    Depthwise Separable Convolution (https://arxiv.org/abs/1610.02357)
    """
    def __init__(self, in_channels, out_channels, kernel_size=(3,3), stride=(1,1), dilation=(1,1), pad='valid'):
        """
        depthwise separable convolution = separate 3*3 conv on each input channel, then use 1*1 conv to map to #out_channels output channels
        :param in_channels:
        :param out_channels:
        :param kernel_size:
        :param stride:
        :param dilation:
        :param pad:
        """
        super().__init__()
        self.conv_depthwise = Conv2D(in_channels=in_channels, out_channels=in_channels, kernel_size=kernel_size, stride=stride, pad=pad, dilation=dilation, num_groups=in_channels)
        self.conv_pointwise = Conv2D(in_channels=in_channels, out_channels=out_channels, kernel_size=(1,1), pad='same')

    def forward(self, x):
        """
        :param x: (B, C, H, W)
        :return:
        """
        self.work_mode = 'train'

        x = self.conv_depthwise.forward(x)   # W = (g, 1, h, w), g = group number = in_channels, (h, w) = kernel_size
        x = self.conv_pointwise.forward(x)   # W = (out_channels, n_channels, 1, 1)
        return x

    def predict(self, x):
        self.work_mode = 'inference'

        x = self.conv_depthwise.predict(x)   # W = (g, h, w), g = group number = in_channels, (h, w) = kernel_size
        x = self.conv_pointwise.predict(x)   # W = (in_channels, out_channels, 1, 1)
        return x


class ShuffleUnit(Module):
    """
    Shuffle unit reference implementation (https://arxiv.org/abs/1610.02357)
    Due to a bug in Theano's convolution, you'll need newer version of Theano
    after this [commit](https://github.com/Theano/Theano/commit/727477b5c3cfb4352f18b07ea4e4cb6df95cc254)
    to run this model.
    """
    def __init__(self, in_channels=256, inner_channels=None, out_channels=None, group_num=4, border_mode='same', batchnorm_mode=1, activation=relu,
                 stride=(1,1), dilation=(1,1), fusion_mode='add'):
        """
        :param in_channels: channel number of block input
        :param inner_channels: channel number inside the block
        :param out_channels: channel number of block output, only used when `fusion_mode` = 'concat', and must > `in_channels`
        :param group_num: number of convolution groups
        :param border_mode: only `same` allowed
        :param batchnorm_mode: {0 | 1 | 2}. 0 means no batch normalization applied; 1 means batch normalization applied to each cnn;
                               2 means batch normalization only applied to the last cnn
        :param activation: default = relu. Note no activation applied to the last element-wise sum output.
        :param stride, dilation: only used for depthwise separable convolution inside
        :param fusion_mode: {'add' | 'concat'}.
        :return y with #channel = `in_channels` when `fusion_mode`='add', #channel = `out_channels` when `fusion_mode`='concat'
        """
        super().__init__()
        self.inner_channels = in_channels//4 if inner_channels is None else inner_channels
        if fusion_mode == 'add':
            self.out_channels = in_channels
        else:
            self.out_channels = out_channels
            if self.out_channels is None or self.out_channels <=in_channels:
                raise ValueError('out_channels must > in_channels')
            self.out_channels -= in_channels
        self.activation     = activation
        self.batchnorm_mode = batchnorm_mode
        self.group_num      = group_num
        self.stride         = as_tuple(stride, 2)
        self.dilation       = as_tuple(dilation, 2)
        self.border_mode    = border_mode
        if border_mode not in {'same'}:
            raise ValueError('Only "same" border mode is supported')
        self.fusion_mode    = fusion_mode
        self.conv1 = Conv2D(in_channels=in_channels, out_channels=self.inner_channels, kernel_size=1, pad=border_mode, num_groups=group_num)
        self.conv2 = DSConv2D(in_channels=self.inner_channels, out_channels=self.inner_channels, kernel_size=3, pad=border_mode, stride=self.stride, dilation=self.dilation)
        self.conv3 = Conv2D(in_channels=self.inner_channels, out_channels=self.out_channels, kernel_size=1, pad=border_mode, num_groups=group_num)
        if batchnorm_mode == 0:   # no batch normalization
            pass
        elif batchnorm_mode == 1: # batch normalization per convolution
            self.bn1 = BatchNorm(input_shape=(None, self.inner_channels, None, None))
            self.bn2 = BatchNorm(input_shape=(None, self.inner_channels, None, None))
            self.bn3 = BatchNorm(input_shape=(None, self.out_channels, None, None))
        elif batchnorm_mode == 2: # only one batch normalization at the end
            self.bn3  = BatchNorm(input_shape=(None, self.out_channels, None, None))
        else:
            raise ValueError('batchnorm_mode should be {0 | 1 | 2}')

    def forward(self, x):
        """
        :param x: (B, C, H, W)
        :return:
        """
        self.work_mode = 'train'

        x0 = x
        x  = self.conv1.forward(x)
        if self.batchnorm_mode == 1:
            x = self.bn1.forward(x)
        x = self.activation(x)          # (B, c*g, H, W)
        x = channel_shuffle(x, self.group_num)
        x  = self.conv2.forward(x)
        if self.batchnorm_mode == 1:
            x = self.bn2.forward(x)
        x = self.activation(x)
        x  = self.conv3.forward(x)
        if self.batchnorm_mode in {1, 2}:
            x = self.bn3.forward(x)
        x = self.activation(x)

        if np.max(self.stride) > 1:
            x0 = pool_2d(x0, ws=(3,3), stride=self.stride, pad=[1,1], ignore_border=True, mode='average_inc_pad')
        if self.fusion_mode == 'add':
            x = x0 + x
        elif self.fusion_mode == 'concat':
            x = tensor.concatenate([x0, x], axis=1)
        else:
            raise ValueError('Only "add" or "concat" fusion mode supported')

        return x

    def predict(self, x):
        self.work_mode = 'inference'

        x0 = x
        x  = self.conv1.predict(x)
        if self.batchnorm_mode == 1:
            x = self.bn1.predict(x)
        x = self.activation(x)          # (B, c*g, H, W)
        x = channel_shuffle(x, self.group_num)
        x  = self.conv2.predict(x)
        if self.batchnorm_mode == 1:
            x = self.bn2.predict(x)
        x = self.activation(x)
        x  = self.conv3.predict(x)
        if self.batchnorm_mode in {1, 2}:
            x = self.bn3.predict(x)
        x = self.activation(x)

        if np.max(self.stride) > 1:
            x0 = pool_2d(x0, ws=(3, 3), stride=self.stride, pad=[1, 1], ignore_border=True, mode='average_inc_pad')
        if self.fusion_mode == 'add':
            x = x0 + x
        elif self.fusion_mode == 'concat':
            x = tensor.concatenate([x0, x], axis=1)
        else:
            raise ValueError('Only "add" or "concat" fusion mode supported')

        return x

class ShuffleUnit_v2(Module):
    """
    Shuffle unit v2 reference implementation (https://arxiv.org/abs/1807.11164)
    """
    def __init__(self, in_channels=256, out_channels=None, border_mode='same', batchnorm_mode=1, activation=relu,
                 stride=1, dilation=1):
        """
        :param in_channels: channel number of block input
        :param out_channels: channel number of block output, only used when `stride` > 1.
        :param border_mode: only `same` allowed
        :param batchnorm_mode: {0 | 1 | 2}. 0 means no batch normalization applied; 1 means batch normalization applied to each cnn;
                               2 means batch normalization only applied to the last cnn
        :param activation: default = relu. Note no activation applied to the last element-wise sum output.
        :param stride, dilation: only used for depthwise separable convolution inside, must be integer scalars
        :return y with #channel = `in_channels` when `fusion_mode`='add', #channel = `out_channels` when `fusion_mode`='concat'
        """
        super().__init__()
        self.activation     = activation
        self.batchnorm_mode = batchnorm_mode
        self.stride         = as_tuple(stride, 2)
        self.dilation       = as_tuple(dilation, 2)
        self.border_mode    = border_mode
        channels = out_channels//2
        assert 2*channels == out_channels, "out_channels must be even"
        if out_channels is None:
            out_channels = in_channels
        if stride == 1:
            assert in_channels == out_channels, 'when stride=1, out_channels must be equal to in_channels'
            self.conv1 = Conv2D(in_channels=channels, out_channels=channels, kernel_size=1, pad=border_mode)
            self.conv2 = DSConv2D(in_channels=channels, out_channels=channels, kernel_size=3, pad=border_mode, dilation=self.dilation)
            self.conv3 = Conv2D(in_channels=channels, out_channels=channels, kernel_size=1, pad=border_mode)
        else:
            self.conv1 = Conv2D(in_channels=in_channels, out_channels=channels, kernel_size=1, pad=border_mode)
            self.conv2 = DSConv2D(in_channels=channels, out_channels=channels, kernel_size=3, pad=border_mode, stride=self.stride, dilation=self.dilation)
            self.conv3 = Conv2D(in_channels=channels, out_channels=channels, kernel_size=1, pad=border_mode)
            self.conv4 = DSConv2D(in_channels=in_channels, out_channels=in_channels, kernel_size=3, pad=border_mode, stride=self.stride, dilation=self.dilation)
            self.conv5 = Conv2D(in_channels=in_channels, out_channels=channels, kernel_size=1, pad=border_mode)

        if batchnorm_mode == 0:  # no batch normalization
            pass
        elif batchnorm_mode == 1:  # batch normalization per convolution
            self.bn1 = BatchNorm(input_shape=(None, channels, None, None))
            self.bn2 = BatchNorm(input_shape=(None, channels, None, None))
            self.bn3 = BatchNorm(input_shape=(None, channels, None, None))
            if stride > 1:
                self.bn4 = BatchNorm(input_shape=(None, in_channels, None, None))
                self.bn5 = BatchNorm(input_shape=(None, channels, None, None))
        elif batchnorm_mode == 2:  # only one batch normalization at the end
            self.bn3 = BatchNorm(input_shape=(None, channels, None, None))
            if stride > 1:
                self.bn5 = BatchNorm(input_shape=(None, channels, None, None))
        else:
            raise ValueError('batchnorm_mode should be {0 | 1 | 2}')

    def forward(self, x):
        """
        :param x: (B, C, H, W)
        :return:
        """
        self.work_mode = 'train'

        if self.stride[0] == 1:
            x1 = x[:, x.shape[1]//2:, :, :]   # the first half channels
            x2 = x[:, :x.shape[1]//2, :, :]   # the last half channels
            x1 = self.conv1.forward(x1)
            if self.batchnorm_mode == 1:
                x1 = self.bn1.forward(x1)
            x1 = self.activation(x1)
            x1 = self.conv2.forward(x1)
            if self.batchnorm_mode == 1:
                x1 = self.bn2.forward(x1)
            x1 = self.conv3.forward(x1)
            if self.batchnorm_mode in {1, 2}:
                x1 = self.bn3.forward(x1)
            x1 = self.activation(x1)
        else:
            x1 = self.conv1.forward(x)
            if self.batchnorm_mode == 1:
                x1 = self.bn1.forward(x1)
            x1 = self.activation(x1)
            x1 = self.conv2.forward(x1)
            if self.batchnorm_mode == 1:
                x1 = self.bn2.forward(x1)
            x1 = self.conv3.forward(x1)
            if self.batchnorm_mode in {1, 2}:
                x1 = self.bn3.forward(x1)
            x1 = self.activation(x1)

            x2 = self.conv4.forward(x)
            if self.batchnorm_mode == 1:
                x2 = self.bn4.forward(x2)
            x2 = self.conv5.forward(x2)
            if self.batchnorm_mode in {1, 2}:
                x2 = self.bn5.forward(x2)
            x2 = self.activation(x2)

        x = tensor.concatenate([x1, x2], axis=1)
        x = channel_shuffle(x, 2)
        return x

    def predict(self, x):
        self.work_mode = 'inference'

        if self.stride[0] == 1:
            x1 = x[:, x.shape[1]//2:, :, :]   # the first half channels
            x2 = x[:, :x.shape[1]//2, :, :]   # the last half channels
            x1 = self.conv1.predict(x1)
            if self.batchnorm_mode == 1:
                x1 = self.bn1.predict(x1)
            x1 = self.activation(x1)
            x1 = self.conv2.predict(x1)
            if self.batchnorm_mode == 1:
                x1 = self.bn2.predict(x1)
            x1 = self.conv3.predict(x1)
            if self.batchnorm_mode in {1, 2}:
                x1 = self.bn3.predict(x1)
            x1 = self.activation(x1)
        else:
            x1 = self.conv1.predict(x)
            if self.batchnorm_mode == 1:
                x1 = self.bn1.predict(x1)
            x1 = self.activation(x1)
            x1 = self.conv2.predict(x1)
            if self.batchnorm_mode == 1:
                x1 = self.bn2.predict(x1)
            x1 = self.conv3.predict(x1)
            if self.batchnorm_mode in {1, 2}:
                x1 = self.bn3.predict(x1)
            x1 = self.activation(x1)

            x2 = self.conv4.predict(x)
            if self.batchnorm_mode == 1:
                x2 = self.bn4.predict(x2)
            x2 = self.conv5.predict(x2)
            if self.batchnorm_mode in {1, 2}:
                x2 = self.bn5.predict(x2)
            x2 = self.activation(x2)

        x = tensor.concatenate([x1, x2], axis=1)
        x = channel_shuffle(x, 2)
        return x

class ShuffleUnit_Stack(Module):
    """
    Stack of shuffle-unit
    downscale factor = (2,2)
    """

    def __init__(self, in_channels, out_channels, inner_channels=None, group_num=4, batchnorm_mode=1, activation=relu, stack_size=3, stride=2, fusion_mode='concat'):
        """
        :param in_channels: channel number of stack input
        :param inner_channels: channel number inside the shuffle-units
        :param out_channels: channel number of stack output, must > `in_channels`
        :param group_num:
        :param batchnorm_mode:
        :param activation:
        :param stack_size:
        :param stride: int or tuple of int, convolution stride for the first unit, default=2
        :param fusion_mode: fusion_mode for the first unit.
        """
        super().__init__()
        self.shuffleunit0 = ShuffleUnit(in_channels=in_channels, inner_channels=inner_channels, out_channels=out_channels, group_num=group_num,
                                        batchnorm_mode=batchnorm_mode, activation=activation, stride=as_tuple(stride, 2), fusion_mode=fusion_mode)
        self.shuffleunit_stack = [self.shuffleunit0]
        for i in range(stack_size):
            shuffleunit = ShuffleUnit(in_channels=out_channels, inner_channels=inner_channels, group_num=group_num,
                                           batchnorm_mode=batchnorm_mode, activation=activation, fusion_mode='add')
            setattr(self, 'shuffleunit%d'%(i+1), shuffleunit)

            self.shuffleunit_stack.append(shuffleunit)

    def forward(self, x):
        self.work_mode = 'train'
        for module in self.shuffleunit_stack:
            x = relu(module.forward(x))
        return x

    def predict(self, x):
        self.work_mode = 'inference'
        for module in self.shuffleunit_stack:
            x = relu(module.predict(x))
        return x

class ShuffleUnit_v2_Stack(Module):
    """
    Stack of shuffle_v2 unit
    downscale factor = (2,2)
    """
    def __init__(self, in_channels, out_channels, batchnorm_mode=1, activation=relu, stack_size=3, stride=2):
        """
        :param in_channels: channel number of stack input
        :param out_channels: channel number of stack output
        :param batchnorm_mode:
        :param activation:
        :param stack_size:
        :param stride: int or tuple of int, convolution stride for the first stack, default=2
        """
        super().__init__()
        self.shuffleunit0 = ShuffleUnit_v2(in_channels=in_channels, out_channels=out_channels, batchnorm_mode=batchnorm_mode,
                                           activation=activation, stride=stride)
        self.shuffleunit_stack = [self.shuffleunit0]
        for i in range(stack_size):
            shuffleunit = ShuffleUnit_v2(in_channels=out_channels, out_channels=out_channels, batchnorm_mode=batchnorm_mode,
                                         activation=activation, stride=1)
            setattr(self, 'shuffleunit%d'%(i+1), shuffleunit)

            self.shuffleunit_stack.append(shuffleunit)

    def forward(self, x):
        self.work_mode = 'train'
        for module in self.shuffleunit_stack:
            x = module.forward(x)
        return x

    def predict(self, x):
        self.work_mode = 'inference'
        for module in self.shuffleunit_stack:
            x = module.predict(x)
        return x

class model_ShuffleNet(Module):
    """
    Model reference implementation of [ShuffleNet](https://arxiv.org/abs/1610.02357) without the final pooling & Dense layer.
    Note no activation applied to the last output
    downscale factor = (32, 32)
    """
    def __init__(self, in_channels, group_num=4, stage_channels=(24, 272, 544, 1088), stack_size=(3, 7, 3), batchnorm_mode=1, activation=relu):
        super().__init__()
        self.conv1 = Conv2D(in_channels=in_channels, out_channels=stage_channels[0], kernel_size=(3,3), stride=(2,2), pad='same')
        self.bn1   = BatchNorm(input_shape=(None, stage_channels[0], None, None))
        self.stage2 = ShuffleUnit_Stack(in_channels=stage_channels[0], out_channels=stage_channels[1], group_num=1, batchnorm_mode=batchnorm_mode,
                                        activation=activation, stack_size=stack_size[0])
        self.stage3 = ShuffleUnit_Stack(in_channels=stage_channels[1], out_channels=stage_channels[2], group_num=group_num, batchnorm_mode=batchnorm_mode,
                                        activation=activation, stack_size=stack_size[1])
        self.stage4 = ShuffleUnit_Stack(in_channels=stage_channels[2], out_channels=stage_channels[3], group_num=group_num, batchnorm_mode=batchnorm_mode,
                                        activation=activation, stack_size=stack_size[2])

    def forward(self, x):
        """
        :param x: (B, C, H, W)
        :return:
        """
        self.work_mode = 'train'

        x = self.conv1.forward(x)
        x = relu(self.bn1.forward(x))
        x = pool_2d(x, ws=(3,3), stride=(2,2), mode='max')
        x = relu(self.stage2.forward(x))
        x = relu(self.stage3.forward(x))
        x = self.stage4.forward(x)
        return x

    def predict(self, x):
        """
        :param x: (B, C, H, W)
        :return:
        """
        self.work_mode = 'inference'
    
        x = self.conv1.predict(x)
        x = relu(self.bn1.predict(x))
        x = pool_2d(x, ws=(3, 3), stride=(2, 2), mode='max')
        x = relu(self.stage2.predict(x))
        x = relu(self.stage3.predict(x))
        x = self.stage4.predict(x)
        return x

class model_ShuffleNet_v2(Module):
    """
    Model reference implementation of [ShuffleNet v2](https://arxiv.org/abs/1807.11164) without the final pooling & Dense layer.
    Note no activation applied to the last output
    """

    def __init__(self, in_channels, stage_channels=(24, 116, 232, 464, 1024), stack_size=(3, 7, 3), batchnorm_mode=1, activation=relu):
        super().__init__()
        self.conv1 = Conv2D(in_channels=in_channels, out_channels=stage_channels[0], kernel_size=(3, 3), stride=(2, 2), pad='same')
        self.bn1   = BatchNorm(input_shape=(None, stage_channels[0], None, None))
        self.stage2 = ShuffleUnit_v2_Stack(in_channels=stage_channels[0], out_channels=stage_channels[1], stack_size=stack_size[0], activation=activation, batchnorm_mode=batchnorm_mode)
        self.stage3 = ShuffleUnit_v2_Stack(in_channels=stage_channels[1], out_channels=stage_channels[2], stack_size=stack_size[1], activation=activation, batchnorm_mode=batchnorm_mode)
        self.stage4 = ShuffleUnit_v2_Stack(in_channels=stage_channels[2], out_channels=stage_channels[3], stack_size=stack_size[2], activation=activation, batchnorm_mode=batchnorm_mode)
        self.conv5  = Conv2D(in_channels=stage_channels[3], out_channels=stage_channels[4], kernel_size=(1,1), pad='same')
        self.bn5    = BatchNorm(input_shape=(None, stage_channels[4], None, None))
    def forward(self, x):
        """
        :param x: (B, C, H, W)
        :return:
        """
        self.work_mode = 'train'

        x = self.conv1.forward(x)
        x = relu(self.bn1.forward(x))
        x = pool_2d(x, ws=(3, 3), stride=(2, 2), mode='max')
        x = self.stage2.forward(x)
        x = self.stage3.forward(x)
        x = self.stage4.forward(x)
        x = self.conv5.forward(x)
        x = self.bn5.forward(x)
        return x

    def predict(self, x):
        """
        :param x: (B, C, H, W)
        :return:
        """
        self.work_mode = 'inference'

        x = self.conv1.predict(x)
        x = relu(self.bn1.predict(x))
        x = pool_2d(x, ws=(3, 3), stride=(2, 2), mode='max')
        x = self.stage2.predict(x)
        x = self.stage3.predict(x)
        x = self.stage4.predict(x)
        x = self.conv5.predict(x)
        x = self.bn5.predict(x)
        return x

class model_ShuffleSeg(Module):
    """
    Model reference implementation of [ShuffleSeg](https://arxiv.org/abs/1803.03816)
    """
    class ShuffleNet(model_ShuffleNet):
        """
        Same with model_ShuffleNet. Output two more heads: feed1 & feed2
        """
        def __init__(self, in_channels, out_channels, group_num=4, stage_channels=(24, 272, 544, 1088), stack_size=(3, 7, 3), batchnorm_mode=1, activation=relu):
            super().__init__(in_channels=in_channels, group_num=group_num, stage_channels=stage_channels,
                             stack_size=stack_size, batchnorm_mode=batchnorm_mode, activation=activation)
            self.conv2 = Conv2D(in_channels=stage_channels[3], out_channels=out_channels, kernel_size=(1,1), pad='same')
       
        def forward(self, x):
            """
             :param x: (B, C, H, W)
             :return:
             """
            self.work_mode = 'train'

            x = self.conv1.forward(x)
            x = relu(self.bn1.forward(x))
            x = pool_2d(x, ws=(3, 3), stride=(2, 2), mode='max')
            x = relu(self.stage2.forward(x))
            feed2 = x
            x = relu(self.stage3.forward(x))
            feed1 = x
            x = relu(self.stage4.forward(x))
            x = self.conv2.forward(x)
            return x, feed1, feed2  # (H, W) = [(7,7), (14, 14), (28, 28)] for (224, 224) input
        
        def predict(self, x):
            """
             :param x: (B, C, H, W)
             :return:
             """
            self.work_mode = 'inference'

            x = self.conv1.predict(x)
            x = relu(self.bn1.predict(x))
            x = pool_2d(x, ws=(3, 3), stride=(2, 2), mode='max')
            x = relu(self.stage2.predict(x))
            feed2 = x
            x = relu(self.stage3.predict(x))
            feed1 = x
            x = relu(self.stage4.predict(x))
            x = self.conv2.predict(x)
            return x, feed1, feed2

    def __init__(self, in_channels=1, Nclass=6, SF_group_num=4, SF_stage_channels=(24, 272, 544, 1088), SF_stack_size=(3, 7, 3), SF_batchnorm_mode=1, SF_activation=relu):
        super().__init__()
        self.encoder = self.ShuffleNet(in_channels=in_channels, out_channels=Nclass, group_num=SF_group_num, stage_channels=SF_stage_channels,
                                       stack_size=SF_stack_size, batchnorm_mode=SF_batchnorm_mode, activation=SF_activation)
        #--- decoder part ---#
        self.convT1  = ConvTransposed2D(in_channels=Nclass, out_channels=Nclass, kernel_size=(4,4), stride=(2,2), pad=(1,1))
        self.bn1     = BatchNorm(input_shape=(None, Nclass, None, None))
        self.conv1   = Conv2D(in_channels=SF_stage_channels[2], out_channels=Nclass, kernel_size=(1,1), pad='same')
        self.bn2     = BatchNorm(input_shape=(None, Nclass, None, None))
        self.convT2  = ConvTransposed2D(in_channels=Nclass, out_channels=Nclass, kernel_size=(4,4), stride=(2,2), pad=(1,1))
        self.bn3     = BatchNorm(input_shape=(None, Nclass, None, None))
        self.conv2   = Conv2D(in_channels=SF_stage_channels[1], out_channels=Nclass, kernel_size=(1,1), pad='same')
        self.bn4     = BatchNorm(input_shape=(None, Nclass, None, None))
        self.convT3  = ConvTransposed2D(in_channels=Nclass, out_channels=Nclass, kernel_size=(16,16), stride=(8,8), pad=(4,4))

    def forward(self, x):
        """
        No activation applied in decoder part, same with the reference paper.
        :param x:
        :return:
        """
        self.work_mode = 'train'

        encoder_out, feed1, feed2 = self.encoder.forward(x)
        upscore2    = self.bn1.forward(self.convT1.forward(encoder_out))
        score_feed1 = self.bn2.forward(self.conv1.forward(feed1))
        fuse_feed1  = score_feed1 + upscore2
        upscore4    = self.bn3.forward(self.convT2.forward(fuse_feed1))
        score_feed2 = self.bn4.forward(self.conv2.forward(feed2))
        fuse_feed2  = score_feed2 + upscore4
        x           = self.convT3.forward(fuse_feed2)
        return x

    def predict(self, x):
        self.work_mode = 'inference'

        encoder_out, feed1, feed2 = self.encoder.predict(x)
        upscore2    = self.bn1.predict(self.convT1.predict(encoder_out))
        score_feed1 = self.bn2.predict(self.conv1.predict(feed1))
        fuse_feed1  = score_feed1 + upscore2
        upscore4    = self.bn3.predict(self.convT2.predict(fuse_feed1))
        score_feed2 = self.bn4.predict(self.conv2.predict(feed2))
        fuse_feed2  = score_feed2 + upscore4
        x           = self.convT3.predict(fuse_feed2)
        return x










# coding:utf-8
'''
Model definition of ResNet
Created   :   7,  6, 2018
Revised   :   7,  6, 2018
All rights reserved
'''
# ------------------------------------------------------------------------------------------------
__author__ = 'dawei.leng'

import theano.tensor as tensor
from ..module import *
from ..functional import *
from ..activation import *


class ResNet_bottleneck(Module):
    """
    [ResNet bottleneck block] (https://arxiv.org/abs/1512.03385).
    """

    def __init__(self, outer_channel=256, inner_channel=64, border_mode='same', batchnorm_mode=1, activation=relu):
        """

        :param outer_channel: channel number of block input
        :param inner_channel: channel number inside the block
        :param border_mode:
        :param batchnorm_mode: {0 | 1 | 2}. 0 means no batch normalization applied; 1 means batch normalization applied to each cnn;
                               2 means batch normalization only applied to the last cnn
        :param activation: default = relu. Note no activation applied to the last element-wise sum output.
        """
        super().__init__()
        self.activation = activation
        self.batchnorm_mode = batchnorm_mode
        self.conv1 = Conv2D(in_channels=outer_channel,  out_channels=inner_channel,  kernel_size=1, pad=border_mode)
        self.conv2 = Conv2D(in_channels=inner_channel,  out_channels=inner_channel,  kernel_size=3, pad=border_mode)
        self.conv3 = Conv2D(in_channels=inner_channel,  out_channels=outer_channel,  kernel_size=1, pad=border_mode)
        if batchnorm_mode == 0:   # no batch normalization
            pass
        elif batchnorm_mode == 1: # batch normalization per convolution
            self.bn1 = BatchNorm(input_shape=(None, inner_channel, None, None))
            self.bn2 = BatchNorm(input_shape=(None, inner_channel, None, None))
            self.bn3 = BatchNorm(input_shape=(None, outer_channel, None, None))
        elif batchnorm_mode == 2: # only one batch normalization at the end
            self.bn3  = BatchNorm(input_shape=(None, outer_channel, None, None))
        else:
            raise ValueError('batchnorm_mode should be {0 | 1 | 2}')

    def forward(self, x):
        """
        :param x: (B, C, H, W)
        :return:
        """
        self.work_mode = 'train'

        x0 = x
        x  = self.conv1.forward(x)
        if self.batchnorm_mode == 1:
            x = self.bn1.forward(x)
        x = self.activation(x)
        x  = self.conv2.forward(x)
        if self.batchnorm_mode == 1:
            x = self.bn2.forward(x)
        x = self.activation(x)
        x  = self.conv3.forward(x)
        if self.batchnorm_mode in {1, 2}:
            x = self.bn3.forward(x)
        x = self.activation(x)
        x = x + x0
        return x

    def predict(self, x):
        self.work_mode = 'inference'

        x0 = x
        x  = self.conv1.predict(x)
        if self.batchnorm_mode == 1:
            x = self.bn1.predict(x)
        x = self.activation(x)
        x  = self.conv2.predict(x)
        if self.batchnorm_mode == 1:
            x = self.bn2.predict(x)
        x = self.activation(x)
        x  = self.conv3.predict(x)
        if self.batchnorm_mode in {1, 2}:
            x = self.bn3.predict(x)
        x = self.activation(x)
        x = x + x0
        return x



# coding:utf-8
# Unit test for GRU class
# Created   :   1, 31, 2018
# Revised   :   1, 31, 2018
# All rights reserved
#------------------------------------------------------------------------------------------------
__author__ = 'dawei.leng'
import os, sys
os.environ['THEANO_FLAGS'] = "floatX=float32, mode=FAST_RUN, warn_float64='raise'"

import theano
from theano import tensor
from dandelion.module import *
from dandelion.activation import *
from lasagne.layers import InputLayer, GRULayer, get_output
import lasagne.nonlinearities as LACT

import dandelion
dandelion_path = os.path.split(dandelion.__file__)[0]
print('dandelion path = %s\n' % dandelion_path)

class build_model_D(Module):
    def __init__(self, in_dim=3, out_dim=3):
        super().__init__()
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.gru = GRU(input_dims=self.in_dim, hidden_dim=self.out_dim, learn_ini=True)
        self.predict = self.forward

    def forward(self, x):
        """

        :param x: (B, T, D)
        :return:
        """
        x = x.dimshuffle((1, 0, 2)) # ->(T, B, D)
        x = self.gru.forward(x, backward=False, only_return_final=False)
        x = x.dimshuffle((1, 0, 2)) # ->(B, T, D)
        # x = tanh(x)
        return x

def build_model_L(in_dim=3, out_dim=3):
    input_var = tensor.ftensor3('x')  # (B, T, D)
    input0 = InputLayer(shape=(None, None, in_dim), input_var=input_var, name='input0')
    gru0 = GRULayer(input0, num_units=out_dim, precompute_input=True,
                      backwards=False, only_return_final=False, learn_init=True,
                      name='gru0')
    return gru0

def test_case_0():
    import numpy as np
    from lasagne_ext.utils import get_layer_by_name

    in_dim, out_dim = 6, 5
    model_D = build_model_D(in_dim=in_dim, out_dim=out_dim)
    model_L = build_model_L(in_dim=in_dim, out_dim=out_dim)

    W_in = np.random.rand(in_dim, 3 * out_dim).astype(np.float32)
    b_in = np.random.rand(3 * out_dim).astype(np.float32)
    W_hid = np.random.rand(out_dim, 3 * out_dim).astype(np.float32)
    h_ini = np.random.rand(out_dim).astype(np.float32)

    model_D.gru.W_in.set_value(W_in)
    model_D.gru.b_in.set_value(b_in)
    model_D.gru.W_hid.set_value(W_hid)
    model_D.gru.h_ini.set_value(h_ini)

    gru_L = get_layer_by_name(model_L, 'gru0')
    gru_L.W_in_to_resetgate.set_value(W_in[:, :out_dim])
    gru_L.W_in_to_updategate.set_value(W_in[:, out_dim:2 * out_dim])
    gru_L.W_in_to_hidden_update.set_value(W_in[:, 2 * out_dim:3 * out_dim])

    gru_L.W_hid_to_resetgate.set_value(W_hid[:, :out_dim])
    gru_L.W_hid_to_updategate.set_value(W_hid[:, out_dim:2 * out_dim])
    gru_L.W_hid_to_hidden_update.set_value(W_hid[:, 2 * out_dim:3 * out_dim])

    gru_L.b_resetgate.set_value(b_in[:out_dim])
    gru_L.b_updategate.set_value(b_in[out_dim:2 * out_dim])
    gru_L.b_hidden_update.set_value(b_in[2 * out_dim:3 * out_dim])

    gru_L.hid_init.set_value(h_ini.reshape((1, out_dim)))

    X = get_layer_by_name(model_L, 'input0').input_var
    y_D = model_D.forward(X)
    y_L = get_output(model_L)

    fn_D = theano.function([X], y_D, no_default_updates=True, on_unused_input='ignore')
    fn_L = theano.function([X], y_L, no_default_updates=True, on_unused_input='ignore')

    for i in range(20):
        x = np.random.rand(2, 5, in_dim).astype(np.float32) - 0.5
        y_D = fn_D(x)
        y_L = fn_L(x)
        diff = np.max(np.abs(y_D - y_L))
        print('i=%d, diff=%0.6f' % (i, diff))
        if diff > 1e-4:
            print('y_D=\n', y_D)
            print('y_L=\n', y_L)
            raise ValueError('diff is too big')


if __name__ == '__main__':

    test_case_0()

    print('Test passed')






# coding:utf-8
# Test for BatchNorm class
# Created   :   2, 27, 2018
# Revised   :   2, 27, 2018
# All rights reserved
#------------------------------------------------------------------------------------------------
__author__ = 'dawei.leng'
import os, sys
os.environ['THEANO_FLAGS'] = "floatX=float32, mode=FAST_RUN, warn_float64='raise'"

import theano
from theano import tensor
from dandelion.module import *
from dandelion.activation import *
from dandelion.objective import *
from dandelion.update import *
from lasagne.layers import InputLayer, BatchNormLayer_DV, get_output, get_all_updates
import lasagne.nonlinearities as LACT
import dandelion
dandelion_path = os.path.split(dandelion.__file__)[0]
print('dandelion path = %s\n' % dandelion_path)

class build_model_D(Module):
    def __init__(self, input_shape=None, axes='auto'):
        super().__init__()
        self.input_shape = input_shape
        self.axes = axes
        self.bn = BatchNorm(input_shape=self.input_shape, axes=self.axes)

    def forward(self, x):
        x = self.bn.forward(x)
        return x

    def predict(self, x):
        return self.bn.predict(x)

def build_model_L(input_shape=None, axes='auto'):
    input_var = tensor.ftensor4('x')
    input0 = InputLayer(shape=input_shape, input_var=input_var, name='input0')
    result = BatchNormLayer_DV(input0, axes=axes, name='bn0')
    return result

def fix_update_bcasts(updates):
    for param, update in updates.items():
        if param.broadcastable != update.broadcastable:
            updates[param] = tensor.patternbroadcast(update, param.broadcastable)
    return updates

def test_case_0():
    import numpy as np
    from lasagne_ext.utils import get_layer_by_name

    B, C, H, W = 2, 1, 8, 8
    input_shape = (None, C, H, W)
    axes = 'auto'

    model_D = build_model_D(input_shape=input_shape, axes=axes)
    model_L = build_model_L(input_shape=input_shape, axes=axes)

    X = get_layer_by_name(model_L, 'input0').input_var
    #--- predict ---#
    if 0:
        y_D = model_D.predict(X)
        y_L = get_output(model_L, deterministic=True)
        fn_L = theano.function([X], y_L, no_default_updates=True)
        fn_D = theano.function([X], y_D, no_default_updates=True)

    #--- train ---#
    if 1:
        y_D = model_D.forward(X)
        y_L = get_output(model_L, deterministic=False)

    update_L = fix_update_bcasts(get_all_updates(model_L))
    update_D = fix_update_bcasts(model_D.collect_self_updates())

    fn_L = theano.function([X], y_L, updates=update_L, no_default_updates=True)
    fn_D = theano.function([X], y_D, updates=update_D, no_default_updates=False)
    # fn_L = theano.function([X], y_L, no_default_updates=True)
    # fn_D = theano.function([X], y_D, no_default_updates=True)


    for i in range(20):
        x = np.random.rand(B, C, H, W).astype(np.float32)
        y_D = fn_D(x)
        y_L = fn_L(x)
        diff = np.sum(np.abs(y_D - y_L))
        print('i=%d, diff=%0.6f' % (i, diff))
        if diff>1e-4:
            print(y_D)
            print(y_L)
            raise ValueError('diff is too big')
    print('test_case_0 passed')

def test_case_1():
    B, C, H, W = 4, 1, 256, 256
    x = tensor.ftensor4('x')
    z = tensor.ftensor4('gt')
    bn   = BatchNorm(input_shape=(B, C, H, W), beta=None, gamma=None)
    conv = Conv2D(in_channels=C, out_channels=C, kernel_size=(3,3), pad='same')
    model = Sequential([bn, conv], activation=relu)
    y = model.forward(x)
    loss = aggregate(squared_error(y, z))
    updates = adadelta(loss, model.collect_params())
    updates.update(model.collect_self_updates())
    # f = theano.function([x], y, no_default_updates=False, updates=fix_update_bcasts(bn.collect_self_updates()))
    f = theano.function([x, z], [y, loss], no_default_updates=False, updates=updates)
    x = np.random.rand(B, C, H, W).astype('float32')
    z = np.random.rand(B, C, H, W).astype('float32')
    y, loss = f(x, z)
    assert y.shape ==(B, C, H, W)
    print('test_case_1 passed')

if __name__ == '__main__':

    # test_case_0()

    test_case_1()

    print('Test passed')